# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['configure']

_composeOptions = (
	'enable',
	'path',
	'systemd.enable',
	'build',
	'start',
)

def configure(env, cfg):
	if not cfg.getboolean('docker', 'enable', fallback = False):
		return
	env.settings.merge(cfg, 'docker', ('enable',))
	for s in cfg.sections():
		if s.startswith('docker-compose:'):
			name = s.replace('docker-compose:', '', 1).strip()
			if name == '':
				raise env.error('docker-compose name is empty')
			if cfg.getboolean(s, 'enable', fallback = True):
				env.settings.merge(cfg, s, _composeOptions)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['configure']

def configure(env, cfg):
	env.settings.merge(cfg, 'service.munin', (
		'config.dir',
		'db.dir',
		'dbdir.user',
		'dbdir.group',
	))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, unlink, makedirs
from shutil import rmtree, copytree, copyfile

__all__ = ['lock', 'unlock', 'fpath', 'create', 'sync', 'copy']

def lock(env):
	bdir = env.build.rootdir()
	env.log("build dir %s" % bdir)
	fn = path.normpath(bdir) + '.lock'
	env.debug("lock %s" % fn)
	env.session.set('lockfn', fn)
	makedirs(path.dirname(bdir), exist_ok = True)
	try:
		fh = open(fn, 'x')
	except FileExistsError:
		raise env.error("lock file exists: %s" % fn)
	fh.write('1')
	fh.flush()
	fh.close()
	_cleandir(env, bdir)

def unlock(env):
	fn = env.session.get('lockfn')
	env.debug("unlock %s" % fn)
	try:
		unlink(fn)
	except FileNotFoundError:
		raise env.error("unlock file not found: %s" % fn)

def _cleandir(env, bdir):
	base = path.normpath(bdir)
	bdirs = [
		bdir,
		base + '.meta',
	]
	bfiles = [
		base + '.zip',
		base + '.env',
		base + '.env.asc',
		base + '.deploy',
	]
	for d in bdirs:
		if path.isdir(d):
			env.debug("rmtree %s" % d)
			rmtree(d) # clean tree
		env.debug("makedirs %s" % d)
		makedirs(d) # recreate base dirs
	for f in bfiles:
		if path.isfile(f):
			env.debug("unlink %s" % f)
			unlink(f)

def _open(env, filename, mode = 'r', meta = False):
	fn = fpath(env, filename, meta = meta)
	if mode != 'r':
		dstdir = path.dirname(fn)
		if not path.isdir(dstdir):
			env.debug("makedirs %s" % dstdir)
			makedirs(dstdir, exist_ok = True)
	env.debug("open(%s) %s" % (mode, fn))
	return open(fn, mode)

def fpath(env, *parts, meta = False):
	bdir = env.build.rootdir()
	if meta:
		bdir = path.normpath(bdir) + '.meta'
	fn = path.join(*parts)
	fn = path.normpath(fn)
	while fn.startswith(path.sep):
		fn = fn.replace(path.sep, '', 1)
	return path.realpath(path.join(bdir, fn))

def create(env, filename, meta = False, **kwargs):
	fh = _open(env, filename, mode = 'x', meta = meta)
	if not meta:
		env.build.addfile(filename, **kwargs)
	return fh

def sync(env, src, dst, **kwargs):
	srcdir = env.assets.rootdir(src)
	dstdir = fpath(env, dst)
	copytree(srcdir, dstdir, symlinks = False)
	env.build.adddir(dst, **kwargs)

def copy(env, src, dst, **kwargs):
	srcfn = env.assets.rootdir(src)
	dstfn = fpath(env, dst)
	dstdir = path.dirname(dstfn)
	if not path.isdir(dstdir):
		makedirs(dstdir)
	copyfile(srcfn, dstfn, follow_symlinks = False)
	env.build.addfile(dst, **kwargs)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path
from pytest import raises

from _sadm.errors import EnvError, ServiceError
from _sadm.plugin import service

def test_defaults(testing_env):
	env = testing_env(name = 'service', profile = 'plugin')
	assert env.name() == 'service'
	assert env.profile() == 'plugin'
	assert env.cfgfile() == path.join('service', 'config.ini')
	assert env.settings.sections() == []

def test_configure(testing_env):
	env = testing_env(name = 'service', profile = 'plugin')
	env.configure()
	assert sorted(env.settings.sections()) == ['sadm', 'service']
	assert env.settings.get('service', 'config.dir') == 'service'
	assert env.settings.getlist('service', 'enable') == ()

def test_enable(testing_env):
	env = testing_env(name = 'service', profile = 'plugin')
	env.configure(cfgfile = path.join('service', 'config-testing.ini'))
	assert env.settings.getlist('service', 'enable') == ('testing',)

def test_enable_error(testing_env):
	env = testing_env(name = 'service', profile = 'plugin')
	with raises(EnvError, match = r'service\.conf file not found'):
		env.configure(cfgfile = path.join('service', 'config-nosvc.ini'))

def test_config_error(testing_env):
	env = testing_env(name = 'service', profile = 'plugin')
	with raises(ServiceError, match = 'Source contains parsing errors'):
		env.configure(cfgfile = path.join('service', 'config-error.ini'))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from glob import glob
from os import path

from _sadm import configure

_expectPlugins = dict(enumerate((
	'testing',
	'sadm',
	'sadmenv',
	'os',
	'os.pkg',
	'os.user',
	'sync',
	'network.iptables',
	'network.fail2ban',
	'vcs.clone',
	'docker',
	'service',
	'service.postfix',
	'service.munin',
	'service.munin_node',
	'service.apache',
)))

def test_plugins_list():
	idx = 0
	for p in configure.pluginsList():
		assert (idx, p) == (idx, _expectPlugins[idx])
		idx += 1

def test_plugins_list_revert():
	idx = max(_expectPlugins.keys())
	for p in configure.pluginsList(revert = True):
		assert (idx, p) == (idx, _expectPlugins[idx])
		idx -= 1

def test_plugins_new():
	known = {}
	for n in _expectPlugins.values():
		known[n] = True
	for p in configure.pluginsList():
		assert known.get(p, False), "new unknown plugin: %s" % p

def test_plugins_list_match_fs():
	pl = list(configure.pluginsList())
	fs = glob(path.join('_sadm', 'plugin', '*', 'config.ini'))
	fs.extend(glob(path.join('_sadm', 'plugin', '**', 'config.ini')))
	fs.extend(glob(path.join('_sadm', 'plugin', '**', '*', 'config.ini')))
	fsok = {}
	for x in fs:
		fsok['.'.join(x.split(path.sep)[2:-1])] = True
	fs = [x for x in fsok.keys()]
	fs.remove('skel')
	fs.append('testing')
	fs = sorted(fs)
	pl = sorted(pl)
	print('-- PL:', pl)
	print('-- FS:', fs)
	assert fs == pl
	assert fs == sorted([n for _, n in _expectPlugins.items()])



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from io import UnsupportedOperation
from os import path, makedirs, unlink
from pytest import raises
from shutil import rmtree

from _sadm.asset import Manager
from _sadm.errors import AssetError, AssetNotFoundError

_rootdir = path.join('tdata', 'testing')

def test_manager():
	rdir = path.realpath(_rootdir)
	m = Manager(_rootdir)
	assert m.rootdir() == rdir
	with m.open('asset.test') as fh:
		fh.close()
	with raises(AssetNotFoundError):
		m.open('nofile')

def test_asset_path():
	rdir = path.realpath(_rootdir)
	m = Manager(_rootdir)
	assert m._path('p0', 'p1', 'p2') == path.join(rdir, 'p0', 'p1', 'p2')
	assert m._path(path.sep, 'p0', 'p1', 'p2') == path.join(rdir, 'p0', 'p1', 'p2')

def test_rootdir():
	rdir = path.realpath(_rootdir)
	m = Manager(_rootdir)
	assert m.rootdir() == rdir
	assert m.rootdir('p0', 'p1', 'p2') == path.join(rdir, 'p0', 'p1', 'p2')

def test_asset_name():
	m = Manager(_rootdir)
	assert m.name('p0', 'p1', 'p2') == path.join('p0', 'p1', 'p2')
	assert m.name(path.sep, path.sep, 'p0', 'p1') == path.join('p0', 'p1')
	assert m.name('.', path.sep, 'p0', 'p1') == path.join('p0', 'p1')
	assert m.name('..', path.sep, 'p0', 'p1') == path.join('p0', 'p1')
	assert m.name('x', '..', '..', '..', 'p0', 'p1') == path.join('p0', 'p1')
	assert m.name('x', '.', 'p0', '..', 'p1') == path.join('x', 'p1')
	assert m.name('x', '..', 'p0', '..', 'p1') == 'p1'
	assert m.name('x', '..', '..', '..', 'p0', '..', '..', 'p1') == 'p1'
	assert m.name('..', '..', '..', 'p0', '..', '..', 'p1') == 'p1'
	assert m.name('.filename') == '.filename'

def test_read_only():
	rdir = path.join('tdata', 'tmp')
	fn = path.join(rdir, 'asset-readonly.test')
	if path.isfile(fn):
		unlink(fn)
	makedirs(rdir, exist_ok = True)
	with open(fn, 'x') as fh:
		fh.write('testing')
	assert path.isfile(fn)
	m = Manager(rdir)
	assert m.isfile('asset-readonly.test')
	with raises(UnsupportedOperation, match = 'not writable'):
		with m.open('asset-readonly.test') as fh:
			fh.write('testing')
	unlink(fn)

def test_oserror():
	rdir = path.join('tdata', 'tmp')
	fn = path.join(rdir, 'asset-oserror.dir')
	if path.isdir(fn):
		rmtree(fn)
	makedirs(fn)
	assert path.isdir(fn)
	m = Manager(rdir)
	assert m.isdir('asset-oserror.dir')
	with raises(AssetError, match = ' Is a directory: '):
		m.open('asset-oserror.dir')
	rmtree(fn)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json
from time import time

from _sadm.errors import SessionError

__all__ = ['Session']

class Session(object):
	_start = None
	_done = False
	_d = None

	def __init__(self):
		self._d = {}

	def start(self):
		if self._start is not None:
			raise SessionError('session already started')
		self._start = time()
		self._done = False

	def _check(self):
		if self._done:
			raise SessionError('session stopped')
		if self._start is None:
			raise SessionError('session not started')

	def stop(self):
		self._check()
		took = time() - self._start
		del self._start
		self._done = True
		return took

	def set(self, opt, val):
		self._check()
		if self._d.get(opt, None) is not None:
			raise SessionError("%s option already set" % opt)
		self._d[opt] = val

	def get(self, opt, default = None):
		self._check()
		if self._d.get(opt, None) is None:
			if default is None:
				raise SessionError("%s option not set" % opt)
			else:
				return default
		return self._d[opt]

	def dump(self, fh, indent = '\t', sort_keys = True):
		self._check()
		json.dump(self._d, fh, indent = indent, sort_keys = sort_keys)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, environ
from subprocess import TimeoutExpired

from _sadm import log, libdir
from _sadm.errors import PluginScriptNotFound, PluginScriptNoExec, PluginScriptTimeout
from _sadm.utils import cmd

__all__ = ['Scripts']

_TTL = 180

class Scripts(object):
	_dir = None
	_env = None

	def __init__(self, pname, dist, sdir = None):
		if sdir is None:
			sdir = libdir.fpath('scripts', dist, pname.replace('.', path.sep))
		self._dir = sdir
		log.debug("%s" % self._dir)
		self._env = environ.copy()

	def run(self, script, args = tuple()):
		spath = path.join(self._dir, script)
		log.debug("run %s %s" % (spath, args))
		s = _Script(spath, args)
		return s.dispatch(env = self._env)

class _Script(object):
	_cmd = None

	def __init__(self, script, args):
		self._cmd = [script]
		self._cmd.extend(args)

	def dispatch(self, env = None):
		if env is None:
			env = environ.copy()
		try:
			return cmd.call(self._cmd, env = env, timeout = _TTL)
		except FileNotFoundError:
			raise PluginScriptNotFound(self._cmd[0])
		except PermissionError:
			raise PluginScriptNoExec(self._cmd[0])
		except TimeoutExpired:
			raise PluginScriptTimeout("%s after %s seconds" % (self._cmd[0], _TTL))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import request
from _sadm import log
from _sadm.web import tpl, syslog
from _sadm.web.app import wapp, view

@wapp.route('/syslog')
@wapp.route('/syslog/last')
@wapp.route('/syslog/last/<limit:int>')
@view('syslog.html')
@tpl.data('syslog')
def index(limit = '100'):
	limit = request.query.get('limit', limit)
	try:
		if int(limit) < 0:
			limit = '0'
	except ValueError:
		limit = '0'
	log.debug("last %s messages" % limit)
	lvlmap = dict()
	for lname, lid in syslog._LVLMAP.items():
		lvlmap[lid] = lname
	return {
		'limit': limit,
		'msgs': syslog.last(int(limit)),
		'lvlmap': lvlmap,
	}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['configure']

def configure(env, cfg):
	if cfg.getboolean('vcs.clone', 'enable', fallback = True):
		for s in cfg.sections():
			if s.startswith('vcs.repo'):
				_configRepo(env, cfg, s)

def _configRepo(env, cfg, sect):
	env.settings.merge(cfg, sect, (
		'type',
		'remote',
		'branch',
		'checkout',
		'path',
		'update',
	))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque
from os import path
from unittest.mock import Mock

class MockTmpFile(object):
	_fn = None

	def __init__(self, suffix = None, prefix = None, dir = None, remove = False):
		if suffix is None:
			suffix = '.mock'
		if prefix is None:
			prefix = __name__
		if dir is None:
			dir = path.join(path.sep, 'tmp')
		self._fn = path.join(dir, prefix + suffix)

	def __enter__(self):
		return self

	def __exit__(self, *args):
		pass

	def close(self):
		pass

	def unlink(self):
		pass

	def write(self, data):
		pass

	def name(self):
		return self._fn

class MockShUtil(object):
	_mock = None
	_expect = None
	_return = None
	_default = None
	makedirs = None
	chmod = None
	chown = None
	mktmp = None
	getcwd = None
	chdir = None

	def __init__(self, cfg):
		self._expect = []
		self._return = {}
		self._default = {}
		self._mock = Mock()
		self.mktmp = self._mock.mock_mktmp
		self.getcwd = self._mock.mock_getcwd
		self.makedirs = self._mock.mock_makedirs
		self.chmod = self._mock.mock_chmod
		self.chown = self._mock.mock_chown
		self.chdir = self._mock.mock_chdir
		self._configure(cfg)

	def _configure(self, cfg):
		if cfg is None:
			return
		self._utilsDefault()
		self._parseConfig(cfg)
		self.mktmp.side_effect = self._mktmp
		self.getcwd.side_effect = self._sideEffect('getcwd')
		self.makedirs.side_effect = self._sideEffect('makedirs')

	def _utilsDefault(self):
		self._default['getcwd'] = path.join(path.sep, 'testing', 'workdir')

	def _parseConfig(self, cfg):
		data = cfg.get('shutil', fallback = '')
		if data != '':
			for l in data.splitlines():
				l = l.strip()
				if l != '':
					x = l.split(';')
					rtrn = x[0].strip()
					cmdline = ';'.join(x[1:]).strip()
					self._expect.append(cmdline)
					util = cmdline.split(' ')[0].strip()
					self._utilReturn(util, rtrn)

	def _utilReturn(self, name, data):
		if name == '':
			raise RuntimeError('mock shutil: util name is empty')
		if self._return.get(name, None) is None:
			self._return[name] = deque()
		self._return[name].appendleft(data)

	def _sideEffect(self, util):
		def wrapper(*args, **kwargs):
			rtrn = self._return.get(util, None)
			if rtrn is None:
				return self._default.get(util, None)
			try:
				data = rtrn.pop()
			except IndexError:
				return self._default.get(util, None)
			if data == '':
				return self._default.get(util, None)
			return data
		return wrapper

	def _mktmp(self, suffix = None, prefix = None, dir = None, remove = False):
		return MockTmpFile(suffix = suffix, prefix = prefix, dir = dir, remove = remove)

	def check(self):
		got = []
		for x in self._mock.mock_calls:
			xname = x[0].replace('mock_', '', 1)
			xargs = x[1]
			cmdline = xname
			if len(xargs) > 0:
				cmdline = "%s %s" % (xname, ' '.join([str(i) for i in xargs]))
			xkwargs = x[2]
			for k, v in xkwargs.items():
				v = str(v)
				cmdline = "%s, %s=%s" % (cmdline, k, v)
			got.append(cmdline)
		assert got == self._expect, \
			"mock shutil\ngot:\n%s\nexpect:\n%s" % ('\n'.join(got), '\n'.join(self._expect))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import tarfile
from os import stat

__all__ = ['deploy']

def deploy(env):
	fn = env.name() + '.tar'
	fpath = env.assets.rootdir(env.name(), fn)
	mtime = stat(fpath).st_mtime
	target = env.settings.get('sadmenv', 'target.dir')
	env.log("%s %s" % (fn, mtime))
	env.log("target %s" % target)
	with open(fpath, 'rb') as fh:
		tar = tarfile.open(fn, 'r:', fileobj = fh)
		for tinfo in tar:
			tinfo.mtime = mtime
			env.log("  %s" % tinfo.name)
			tar.extract(tinfo, path = target, set_attrs = True)
		tar.close()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, unlink
from pytest import raises

from _sadm.deploy import extractor
from _sadm.errors import BuildError

extractorfn = path.join('tdata', 'build', 'envsetup', 'testing.deploy')
envfn = path.join('tdata', 'build', 'envsetup', 'testing.env')
zipenvfn = path.join('tdata', 'build', 'envsetup', 'testing.zip')

def test_gen(env_setup):
	env = env_setup(action = 'build')
	assert path.isfile(extractorfn)
	assert path.isfile(envfn)
	assert path.isfile(zipenvfn)

def test_gen_error(env_setup):
	env = env_setup(action = 'build')
	with raises(BuildError, match = "%s file exists" % extractorfn):
		extractor.gen(env)

def test_env_error(env_setup):
	env = env_setup(action = 'build')
	assert path.isfile(extractorfn)
	assert path.isfile(envfn)
	unlink(extractorfn)
	unlink(envfn)
	with raises(BuildError, match = "%s file not found" % envfn):
		extractor.gen(env)

def test_zipenv_error(env_setup):
	env = env_setup(action = 'build')
	assert path.isfile(extractorfn)
	assert path.isfile(zipenvfn)
	unlink(extractorfn)
	unlink(zipenvfn)
	with raises(BuildError, match = "%s file not found" % zipenvfn):
		extractor.gen(env)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque

from _sadm.utils import path
from _sadm.utils.sh import chdir, getcwd

__all__ = ['check', 'getRepo']

_OPTIONS = (
	('type', 'git'),
	('remote', ''),
	('branch', 'master'),
	('checkout', ''),
	('path', ''),
	('update', False),
)

def check(env):
	status = deque()
	repos = _getRepos(env)
	for name, repo in repos.items():
		typ = repo['type']
		env.log("check %s repo %s" % (typ, name))
		rpath = path.abspath(repo['path'])
		if not path.isdir(rpath):
			env.error("repo %s dir %s not found" % (name, rpath))
			status.append(('MISS', name, typ, repo))
			continue
		oldwd = getcwd()
		try:
			chdir(rpath)
			env.log("repo %s dir %s" % (name, rpath))
			if typ == 'git':
				_gitCheck(env, status, name, repo)
		except Exception as err:
			raise env.error("%s" % err)
		finally:
			chdir(oldwd)
		status.append(('OK', name, typ, repo))
	return status

def _getRepos(env):
	repos = {}
	for sect in env.settings.sections():
		if sect.startswith('vcs.repo:'):
			rn = sect.replace('vcs.repo:', '', 1)
			repos[rn] = {} # override if already defined
			for key, defval in _OPTIONS:
				if isinstance(defval, bool):
					repos[rn][key] = env.settings.getboolean(sect, key, fallback = defval)
				else:
					repos[rn][key] = env.settings.get(sect, key, fallback = defval)
	return repos

def _gitCheck(env, status, name, repo): # TODO!!
	pass



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm.configure import pluginsList, getPlugin
from _sadm.env.settings import Settings

# load plugins
import _sadm.plugin.load

__all__ = ['configure']

_cfgfilter = {}

def configure(env, cfgfile = None):
	if cfgfile is None:
		cfgfile = env.cfgfile()
	fn = cfgfile
	env.log("%s" % fn)
	env.start('configure')
	_getcfg(env, fn)
	env.end('configure')

def _getcfg(env, fn):
	runconfigure = True
	cfg = Settings()
	with env.assets.open(fn) as fh:
		cfg.read_file(fh)
	n = cfg.get('sadmenv', 'name', fallback = None)
	if n is None:
		# env config
		n = cfg.get('sadm', 'env', fallback = None)
	else:
		# deploy mode
		runconfigure = False
	# check env/config names match
	if n != env.name():
		raise env.error("%s invalid config name from %s file" % (n, env.assets.rootdir(fn)))
	if runconfigure:
		# env config
		_load(env, cfg)
	else:
		# deploy mode
		with env.assets.open(fn) as fh:
			env.settings.read_file(fh)

def _load(env, cfg, forcePlugins = None):
	env.debug("registered plugins %s" % ','.join([p for p in pluginsList()]))
	if forcePlugins is None:
		forcePlugins = {}
		for p in env.config.listPlugins(env.profile()):
			forcePlugins[p] = True
	env.debug("plugins force enable: %s" % ','.join([p for p in forcePlugins.keys()]))
	for p in pluginsList():
		ena = cfg.has_section(p)
		forceEna = forcePlugins.get(p, False)
		if ena or forceEna:
			env.debug("%s plugin enabled" % p)
			_pluginConfigure(env, cfg, p)
		else:
			env.debug("%s plugin disabled" % p)

def _pluginConfigure(env, cfg, n):
	p = getPlugin(n, 'configure')
	_initPlugin(env, p, cfg)
	env.log(p.name)
	p.mod.configure(env, cfg)
	if hasattr(p.mod, 'cfgfilter'):
		env.debug("plugin %s register cfgfilter" % p.name)
		_cfgfilter[p.name] = getattr(p.mod, 'cfgfilter')

def _initPlugin(env, p, cfg):
	env.debug("init %s" % p.config)
	pcfg = Settings()
	with open(p.config, 'r') as fh:
		pcfg.read_file(fh)
	for dep in pcfg.plugins():
		pfilter = _cfgfilter.get(dep, None)
		if pfilter is not None:
			pfilter(env, pcfg, dep)
	if not env.settings.has_section(p.name):
		env.settings.add_section(p.name)
	for opt, val in pcfg.items(p.name, raw = True):
		env.settings[p.name][opt] = val



# coding: utf-8
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import _sadm.version

# -- Project information -----------------------------------------------------

project = 'sadm'
copyright = '2019, Jeremías Casteglione'
author = 'Jeremías Casteglione'

# The full version, including alpha/beta/rc tags
release = _sadm.version.get()

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
	'sphinx.ext.autodoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

source_suffix = {
	'.rst': 'restructuredtext',
	'.md': 'markdown',
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'traditional'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from .debian import deploy as debian

__all__ = ['deploy']

def deploy(env):
	dn = env.dist()
	if dn == 'debian':
		debian.deploy(env)
	else:
		env.error("unsupported dist %s" % dn)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['BitbucketProvider']

class BitbucketProvider(object):

	def auth(self, req, cfg): # FIXME
		pass

	def repoArgs(self, obj, cfg):
		return {
			'repo.path': cfg.get('path'),
		}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
from os import path

tlib = path.join(path.dirname(path.dirname(__file__)), 'tlib')
sys.path.insert(0, tlib)

from _sadmtest.conf import *



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import builddir

from .configure import getInfo

__all__ = ['build']

def build(env):
	dn = env.dist() + '.'
	for opt in env.settings['sync']:
		if not opt.startswith(dn):
			continue
		inf = getInfo(env.settings, opt)
		_sync(env, inf['cfg.opt'], inf['src'], inf['dst'], inf['args'])
	for inf in env.session.get('sync'):
		_sync(env, inf['cfg.opt'], inf['src'], inf['dst'], inf['args'])

def _sync(env, opt, src, dst, args):
	if env.assets.isdir(src):
		env.log("%s dir %s -> %s" % (opt, src, dst))
		builddir.sync(env, src, dst, **args)
	else:
		env.log("%s file %s -> %s" % (opt, src, dst))
		with env.assets.open(src) as sfh:
			with builddir.create(env, dst, **args) as dfh:
				dfh.write(sfh.read())



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import systemd, path, sh, cmd

__all__ = ['deploy']

def deploy(env):
	systemd.restart('docker')
	for s in env.settings.sections():
		if s.startswith('docker-compose:'):
			name = s.replace('docker-compose:', '', 1).strip()
			cfg = env.settings[s]
			_composeBuild(env, name, cfg)
			_composeConfigure(env, name, cfg)

def _composeBuild(env, name, cfg):
	env.log("docker-compose build %s" % name)
	if cfg.getboolean('build', fallback = True):
		workdir = cfg.get('path')
		oldwdir = sh.getcwd()
		try:
			sh.chdir(workdir)
			cmd.callCheck(['docker-compose', 'build'])
		finally:
			sh.chdir(oldwdir)

def _composeConfigure(env, name, cfg):
	env.log("docker-compose configure %s" % name)
	enable = cfg.getboolean('systemd.enable', fallback = True)
	service = "docker-compose-%s.service" % name
	if enable:
		systemd.enable(service)
		if cfg.getboolean('start', fallback = True):
			systemd.start(service)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import os
import shutil as _sh
import tempfile

__all__ = ['makedirs', 'chmod', 'chown', 'mktmp', 'getcwd', 'chdir']

class TmpFile(object):
	_fd = None
	_fn = None
	_rm = None

	def __init__(self, suffix = None, prefix = None, dir = None, remove = False):
		if prefix is not None:
			if not prefix.endswith('.'):
				prefix = "%s." % prefix
		self._fd, self._fn = tempfile.mkstemp(suffix = suffix,
			prefix = prefix, dir = dir, text = False)
		self._rm = remove

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()
		if self._rm:
			self.unlink()

	def close(self):
		os.close(self._fd)

	def unlink(self):
		os.unlink(self._fn)

	def write(self, data):
		if isinstance(data, str):
			data = data.encode('utf-8')
		os.write(self._fd, data)

	def name(self):
		return self._fn

class _ShUtil(object):
	makedirs = os.makedirs
	chmod = os.chmod
	chown = _sh.chown
	mktmp = None
	getcwd = os.getcwd
	chdir = os.chdir

	def __init__(self):
		self.mktmp = self._mktmp

	def _mktmp(self, suffix = None, prefix = None, dir = None, remove = False):
		return TmpFile(suffix = suffix, prefix = prefix, dir = dir, remove = remove)

shutil = _ShUtil()

def makedirs(name, mode = 0o0755, exists_ok = False):
	return shutil.makedirs(name, mode = mode, exist_ok = exists_ok)

def chmod(path, mode):
	return shutil.chmod(path, mode)

def chown(path, user = None, group = None):
	return shutil.chown(path, user = user, group = group)

def mktmp(suffix = None, prefix = None, dir = None, remove = False):
	return shutil.mktmp(suffix = suffix, prefix = prefix, dir = dir, remove = remove)

def getcwd():
	return shutil.getcwd()

def chdir(path):
	return shutil.chdir(path)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['cfgfilter', 'configure']

_options = ('update', 'install', 'remove', 'prune', 'common.enable')

def cfgfilter(env, pcfg, dep):
	dn = env.dist() + '.'
	def pfilter(opt):
		if opt.startswith(dn):
			for n in _options:
				n = '.' + n
				if opt.endswith(n):
					return opt
		return None
	env.settings.merge(pcfg, dep, pfilter)

def configure(env, cfg):
	env.settings.merge(cfg, 'os.pkg', _options)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sqlite3

from bottle import __version__ as bottle_version
from datetime import date
from sys import version as python_version

from _sadm import log
from _sadm.web import tpl
from _sadm.web.app import wapp, view

@wapp.route('/about')
@view('about.html')
@tpl.data('about')
def about():
	log.debug("about")
	return {
		'pythonVersion': python_version,
		'bottleVersion': bottle_version,
		'sqliteVersion': "%s" % sqlite3.version,
		'sqliteLibVersion': "%s" % sqlite3.sqlite_version,
		'curyear': str(date.today().year),
	}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import builddir

__all__ = ['build']

def build(env):
	_setHostname(env)

def _setHostname(env):
	hostname = env.settings.get('os', 'hostname')
	hfn = env.settings.get('os', 'hostname.file')
	env.debug("hostname file %s" % hfn)
	with builddir.create(env, hfn) as fh:
		fh.write(hostname.strip())
		fh.write('\n')
		fh.flush()
	env.log("%s: %s" % (hfn, hostname))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import builddir

def build(env):
	env.log('build')
	testing_error = env.settings.get('testing', 'testing.error', fallback = '')
	if testing_error == 'env_session_error':
		env.session.start()
	with builddir.create(env, 'sadm.testing') as fh:
		fh.write('testing\n')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import pytest
from os import path, makedirs, unlink
from shutil import rmtree, move

__all__ = [
	'env_setup',
	'testing_env',
	'testing_plugin',
	'testing_profile',
	'testing_settings',
]

# logger

from _sadm import log
log._colored = False
log.init('quiet')

# remove version autogen module

for n in ('_version.py', '_version_build.py'):
	fn = path.join('_sadm', n)
	if path.isfile(fn):
		unlink(fn)

# register testing plugins

import _sadmtest.plugin.testing

# config

from _sadm import cfg
cfg._cfgFile = path.join('tdata', 'sadm.cfg')

# testing profile

from _sadm.env import profile

@pytest.fixture
def testing_profile():
	def wrapper(name = 'testing'):
		return profile.Profile(name)
	return wrapper

# testing env

from _sadm import env

@pytest.fixture
def testing_env():
	def wrapper(name = 'testing', profile = 'testing'):
		e = env.Env(profile, name)
		_cleanEnv(e)
		return e
	return wrapper

# testing settings

@pytest.fixture
def testing_settings():
	envmod = env
	def wrapper(profile = 'testing', env = 'testing', cfgfile = None):
		if cfgfile is not None:
			cfgfile = path.join(profile, cfgfile)
		e = envmod.Env(profile, env)
		_cleanEnv(e)
		e.configure(cfgfile = cfgfile)
		return e.settings
	return wrapper

# testing env setup

from _sadm.env import action as envAction

@pytest.fixture
def env_setup():
	return _newEnv

def _newEnv(name = 'testing', profile = 'envsetup', configure = False,
	cfgfile = None, action = None, mkdirs = False, config = None):
	if action is not None:
		# configure is called from cmd.run
		configure = False
	e = env.Env(profile, name, config = config)
	_cleanEnv(e, mkdirs = mkdirs)
	if configure:
		e.configure(cfgfile = cfgfile)
	if action is not None:
		envAction.run(e, action)
	return e

def _cleanEnv(env, mkdirs = False):
	tmpdir = path.join('tdata', 'tmp')
	bdir = path.normpath(env.build.rootdir())
	pdir = path.realpath(path.join('tdata', env.profile(), env.name()))
	deploydir = path.join(path.dirname(path.dirname(bdir)), 'deploy', env.name())
	_dirs = (
		tmpdir,
		bdir,
		bdir + '.meta',
		deploydir,
	)
	for d in _dirs:
		if path.isdir(d):
			rmtree(d)
	_files = (
		bdir + '.zip',
		bdir + '.env',
		bdir + '.env.asc',
		bdir + '.deploy',
		bdir + '.lock',
		path.join(pdir, '.lock'),
	)
	for f in _files:
		if path.isfile(f):
			unlink(f)
	if mkdirs:
		for d in _dirs:
			makedirs(d)
	else:
		makedirs(tmpdir)

# testing plugins

from _sadm.utils import systemd
systemd._ctlcmd = ['systemctl']

from _sadmtest.plugin import Plugin

@pytest.fixture
def testing_plugin():
	return _newPlugin

def _newPlugin(name = 'testing', ns = '_sadm', cfgfn = None,
	deploy = False, buildDeploy = True):
	pdir = name.replace('.', path.sep)
	profile = 'plugin'
	config = None
	if deploy:
		profile = 'deploy'
		config = cfg.new(cfgfile = path.join('tdata', 'deploy.cfg'))
	env = _newEnv(profile = profile, name = name, config = config)
	if cfgfn is not None:
		env._cfgfile = path.join(pdir, 'config', cfgfn)
	p = Plugin(name, env, ns = ns)
	if deploy and buildDeploy:
		_buildDeploy(name, ns = ns)
	return p

def _buildDeploy(pname, ns = '_sadm'):
	print('-- deploy plugin build:', pname)
	depdir = path.join('tdata', 'deploy', 'plugin', pname)
	targetdir = path.join('tdata', 'deploy.target', 'plugin', pname)
	for dirrm in (depdir, targetdir):
		if path.isdir(dirrm):
			rmtree(dirrm)
	p = _newPlugin(pname, ns = ns, cfgfn = 'build.ini')
	p.build()
	makedirs(depdir, exist_ok = True)
	metadir = path.join('tdata', 'build', 'plugin', pname + '.meta')
	for fn in ('configure.ini', 'meta.json', pname + '.tar'):
		src = path.join(metadir, fn)
		dst = path.join(depdir, fn)
		move(src, dst)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import systemd
from _sadm.utils.cmd import call, callCheck
from _sadm.utils.sh import mktmp

__all__ = ['deploy']

def deploy(env):
	_autoconf(env)
	systemd.stop('munin-node')
	systemd.start('munin-node')
	systemd.status('munin-node')

def _autoconf(env):
	env.log('autoconf')
	tmpfh = mktmp(prefix = __name__)
	tmpfh.write('exit 128')
	tmpfh.close()
	tmpfn = tmpfh.name()
	env.debug("tmpfn %s" % tmpfn)
	try:
		call("munin-node-configure --shell >%s 2>&1" % tmpfn)
		callCheck("/bin/sh -eu %s" % tmpfn)
	finally:
		tmpfh.unlink()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
from _sadm import log, env
from _sadm.cmd import flags

def _getArgs():
	parser = flags.new('sadm-build', desc = 'build sadm env')
	return flags.parse(parser)

def main():
	args = _getArgs()
	log.debug("build %s/%s" % (args.profile, args.env))
	rc, _ = env.run(args.profile, args.env, 'build')
	return rc

if __name__ == '__main__':
	sys.exit(main())



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm.errors import AssetError, AssetNotFoundError

class Manager(object):
	_dir = None

	def __init__(self, rootdir = '.'):
		self._dir = path.realpath(path.normpath(rootdir))

	def rootdir(self, *parts):
		if len(parts) > 0:
			return self._path(*parts)
		return self._dir

	def name(self, *parts):
		strip = True
		relname = path.normpath(path.join(*parts))
		while strip:
			strip = False
			for prefix in (path.sep, '..', '.' + path.sep):
				if relname.startswith(prefix):
					relname = relname[len(prefix):]
					strip = True
		return relname

	def _path(self, name, *parts):
		return path.join(self._dir, self.name(name, *parts))

	def open(self, name, *parts):
		try:
			return open(self._path(name, *parts), 'r')
		except FileNotFoundError as err:
			raise AssetNotFoundError(str(err))
		except OSError as err:
			raise AssetError(str(err))

	def isdir(self, name, *parts):
		n = self._path(name, *parts)
		return path.exists(n) and path.isdir(n)

	def isfile(self, name, *parts):
		n = self._path(name, *parts)
		return path.exists(n) and path.isfile(n)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.env.settings import Settings

__all__ = ['configure']

def configure(env, cfg):
	env.settings.merge(cfg, 'os', (
		'users.config.dir',
		'users.config.file',
		'users.dirmode',
		'users.filemode',
	))

	udir = env.settings.get('os', 'users.config.dir')
	fn = env.settings.get('os', 'users.config.file')

	db = Settings()
	with env.assets.open(udir, fn) as fh:
		env.debug("users config %s" % fh.name)
		db.read_file(fh)

	_addgroups(env, db, cfg)
	_addusers(env, db, cfg)

def _addgroups(env, db, cfg):
	l = cfg.getlist('os.group', 'add', fallback = [])
	for group in l:
		gid = db['groups'][group]
		env.debug("add group %d %s" % (int(gid), group))
		if not env.settings.has_section('os.group'):
			env.settings.add_section('os.group')
		env.settings['os.group'][group] = gid

def _addusers(env, db, cfg):
	l = cfg.getlist('os.user', 'add', fallback = [])
	for user in l:
		uid = db['users'][user]
		env.debug("add user %d %s" % (int(uid), user))
		env.settings['os.user'][user] = uid
		env.settings.add_section("os.user.%s" % user)
		for opt, val in db["user.%s" % user].items():
			env.settings["os.user.%s" % user][opt] = val



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque
from os import devnull
from pwd import getpwnam

from _sadm.utils.cmd import call

__all__ = ['check']

def check(env):
	diff = deque()
	_checkGroups(diff, env)
	_checkUsers(diff, env)
	return diff

def _checkGroups(diff, env):
	for group in env.settings['os.group']:
		gid = env.settings.getint('os.group', group)
		rc = call("getent group %s >%s" % (group, devnull))
		if rc == 2:
			diff.append(('group', group, gid))
			env.warn("%d %s not found" % (gid, group))
		elif rc == 0:
			env.log("%d %s OK" % (gid, group))
		else:
			raise env.error("getent group command failed: %d" % rc)

def _checkUsers(diff, env):
	for user in env.settings['os.user']:
		uid = env.settings.getint('os.user', user)
		try:
			info = getpwnam(user)
		except KeyError:
			diff.append(('user', user, uid))
			env.warn("%d %s not found" % (uid, user))
		else:
			if info.pw_uid != uid:
				env.warn("%d %s uid %d does not match" % (uid, user, info.pw_uid))
			else:
				env.log("%d %s OK" % (uid, user))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json

from base64 import b64encode
from os import path, chmod

from _sadm import version, libdir
from _sadm.errors import BuildError

__all__ = ['gen']

def gen(env):
	_vars = {
		'env': env.name(),
		'rootdir': path.join(path.sep, 'opt', 'sadm'),
	}
	try:
		env.log("%s.deploy" % env.name())
		base = path.normpath(env.build.rootdir())
		fn = base + '.deploy'
		_write(fn, _getcargo(env, base), _vars)
	except FileExistsError:
		raise BuildError("%s file exists" % fn)

def _getcargo(env, base):
	cargo = {}
	for ext in ('.zip', '.env', '.env.asc'):
		name = base + ext
		if path.isfile(name):
			env.log("load %s" % path.basename(name))
			cargo[env.name() + ext] = _load(name)
		else:
			if ext != '.env.asc':
				raise BuildError("%s file not found" % name)
	return cargo

def _load(fn):
	with open(fn, 'rb') as fh:
		return b64encode(fh.read()).decode()

def _write(fn, cargo, _vars):
	sk = True
	indent = '\t'
	with libdir.fopen('deploy', 'self_extract.py') as src:
		with open(fn, 'x', encoding = 'utf-8') as fh:
			for line in src.readlines():
				if line.startswith('_cargo'):
					fh.write("_cargo = %s\n" % json.dumps(cargo,
						indent = indent, sort_keys = sk))
				elif line.startswith('_vars'):
					fh.write("_vars = %s\n" % json.dumps(_vars,
						indent = indent, sort_keys = sk))
				else:
					fh.write(line)
			fh.write("\n# sadm version %s (build %s)\n" % (version.get(),
				version.build()))
			fh.flush()
	chmod(fn, 0o0500)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json

from pytest import raises
from os import path, unlink, makedirs
from time import time

from _sadm import asset
from _sadm.env import Env, _lock, _unlock
from _sadm.env import run as envrun
from _sadm.env.profile import Profile
from _sadm.env.settings import Settings
from _sadm.errors import EnvError, AssetNotFoundError, SessionError

def test_env(testing_env):
	e = testing_env()
	assert isinstance(e, Env)
	assert e.name() == 'testing'
	assert e.profile() == 'testing'
	assert e.cfgfile() == path.join('testing', 'config.ini')
	assert isinstance(e._profile, Profile)
	assert e._profile.name() == 'testing'
	assert isinstance(e.assets, asset.Manager)
	assert isinstance(e.settings, Settings)

def test_env_configure(testing_env):
	e = testing_env()
	assert isinstance(e.settings, Settings)
	with raises(AssetNotFoundError, match = 'No such file or directory'):
		e._cfgfile = 'noconfig.ini'
		e.configure()

def test_env_error(testing_env):
	with raises(EnvError, match = 'env not found'):
		Env('testing', 'noenv')

def test_load_error(testing_env):
	e = testing_env()
	with raises(EnvError, match = 'config file not set'):
		e._load(fn = '')
	with raises(EnvError, match = 'noprofile.dir directory not found'):
		e._load(pdir = 'tdata/noprofile.dir')
	with raises(EnvError, match = 'profile-dir-isfile.test is not a directory'):
		e._load(pdir = 'tdata/profile-dir-isfile.test')

def test_start_end_action(testing_env):
	e = testing_env()
	assert [n for n in e._run.keys()] == []
	e.start('testing_action')
	assert [n for n in e._run.keys()] == ['testing_action']
	assert e._run['testing_action'].get('start', '') != ''
	assert e._run['testing_action'].get('tag.prev', None) is not None
	assert e._run['testing_action'].get('end', None) is None
	e.end('testing_action')
	assert e._run['testing_action'].get('end', '') != ''

def test_start_end_error(testing_env):
	e = testing_env()
	assert [n for n in e._run.keys()] == []
	e.start('testing_error')
	with raises(EnvError, match = 'testing_error action already started'):
		e.start('testing_error')
	assert [n for n in e._run.keys()] == ['testing_error']
	with raises(EnvError, match = 'testing_end_error action was not started'):
		e.end('testing_end_error')

def test_report(testing_env):
	e = testing_env()
	e.start('action1')
	e.end('action1')
	e.report('testing_report')
	e.report('testing_report', startTime = time() - 10)
	e.start('action2')
	with raises(EnvError, match = r'not finished action\(s\): action2'):
		e.report('testing_report')

def test_lock(testing_env):
	fn = path.join('tdata', 'testing', '.lock')
	e = testing_env()
	with e.lock():
		assert e._lockfn.endswith(fn)
		assert path.isfile(fn)
	assert not path.isfile(fn)
	assert e._lockfn is None

def test_lock_error(testing_env):
	fn = path.join('tdata', 'testing', '.lock')
	e = testing_env()
	_lock(e)
	assert path.isfile(fn)
	with raises(EnvError, match = 'lock file exists:'):
		_lock(e)
	_unlock(e)
	assert not path.isfile(fn)

def test_unlock_error(testing_env):
	fn = path.join('tdata', 'testing', '.lock')
	e = testing_env()
	_lock(e)
	assert path.isfile(fn)
	assert e._lockfn is not None
	_unlock(e)
	assert not path.isfile(fn)
	assert e._lockfn is None
	_unlock(e)
	assert e._lockfn is None
	e._lockfn = fn
	e.session.start()
	with raises(EnvError, match = 'unlock file not found:'):
		_unlock(e)
	assert not path.isfile(fn)

def test_env_nodir_error():
	rc, _ = envrun('testing', 'testing.nodir', 'build')
	assert rc == 1

def test_run():
	cfgfn = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'configure.ini')
	if path.isfile(cfgfn):
		unlink(cfgfn)
	assert not path.isfile(cfgfn)
	rc, _ = envrun('envsetup', 'testing', 'build')
	assert rc == 0
	assert path.isfile(cfgfn)
	unlink(cfgfn)

def test_run_env_error():
	cfgfn = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'configure.ini')
	if path.isfile(cfgfn):
		unlink(cfgfn)
	assert not path.isfile(cfgfn)
	rc, err = envrun('envsetup', 'testing', 'configure')
	assert err == EnvError
	assert rc == 1
	assert not path.isfile(cfgfn)

def test_run_error():
	lockfn = path.join('tdata', 'build', 'envsetup', 'testing.errors.lock')
	if path.isfile(lockfn):
		unlink(lockfn)
	rc, err = envrun('envsetup', 'testing.errors', 'build')
	assert err == SessionError
	assert rc == 2

def test_envsetup(env_setup):
	env = env_setup(configure = True)
	assert env.name() == 'testing'
	assert env.profile() == 'envsetup'
	assert env.session._start is not None # env.configure() was run

def test_envsetup_action(env_setup):
	mdir = path.join('tdata', 'build', 'envsetup', 'testing')
	zfn = path.join('tdata', 'build', 'envsetup', 'testing.zip')
	assert not path.isdir(mdir)
	assert not path.isfile(zfn)
	env_setup(action = 'build')
	assert path.isdir(mdir)
	assert path.isfile(zfn)
	with raises(EnvError, match = 'invalid action configure'):
		env_setup(action = 'configure')
	assert not path.isdir(mdir)

def test_env_default(testing_env):
	e = testing_env(name = 'default', profile = 'default')
	assert e.name() == 'testing'
	assert e.profile() == 'testing'

def test_build_checksum(env_setup):
	metafn = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'meta.json')

	# cksum0
	env = env_setup(action = 'build')
	assert path.isfile(metafn)
	with open(metafn, 'r') as fh:
		data = json.load(fh)
	cksum0 = data.get('sadm.env.checksum', None)

	unlink(metafn)
	assert not path.isfile(metafn)

	# cksum1
	env = env_setup(action = 'build')
	assert path.isfile(metafn)
	with open(metafn, 'r') as fh:
		data = json.load(fh)
	cksum1 = data.get('sadm.env.checksum', None)

	assert cksum1 == cksum0



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm.deploy import extractor, self_extract

deployfn = path.join('tdata', 'build', 'envsetup', 'build.test.deploy')
rootdir = path.join('tdata', 'tmp', 'self_extract.test')

def test_extract(env_setup):
	env = env_setup(name = 'build.test', action = 'build')
	assert path.isfile(deployfn)
	assert self_extract._vars == {}
	assert self_extract._cargo == {}
	self_extract._vars = {
		'env': 'build.test',
		'rootdir': rootdir,
	}
	self_extract._cargo = extractor._getcargo(env,
		path.normpath(env.build.rootdir()))
	assert sorted(self_extract._cargo.keys()) == ['build.test.env', 'build.test.zip']
	dstdir = path.join(rootdir, 'env')
	assert not path.isdir(dstdir)
	self_extract.extract(dstdir)
	assert path.isdir(dstdir)
	assert path.isfile(path.join(dstdir, 'build.test.env'))
	assert path.isfile(path.join(dstdir, 'build.test.zip'))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import bottle

from _sadm import libdir

__all__ = ['wapp', 'view']

_cfgfn = libdir.fpath('web', 'app.conf')
_htmldir = libdir.fpath('web', 'html')
_staticdir = libdir.fpath('web', 'static')

bottle.TEMPLATE_PATH.insert(0, _htmldir)
view = bottle.view

wapp = bottle.Bottle()
wapp.config.load_config(_cfgfn)

@wapp.route('/static/<filename:path>')
def _static(filename):
	return bottle.static_file(filename, root = _staticdir, download = False)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log
from _sadm.listen import wapp

__all__ = ['start']

def start():
	w = wapp.init()
	quiet = log.curLevel() == 'quiet'
	debug = wapp.config.getboolean('sadm.listen', 'debug', fallback = False)
	host = wapp.config.get('sadm.listen', 'host', fallback = '127.0.0.1')
	port = wapp.config.getint('sadm.listen', 'port', fallback = 3666)
	return w.run(host = host, port = port, debug = debug,
		reloader = debug, quiet = quiet)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import builddir, path

__all__ = ['build']

def build(env):
	jaildir = env.assets.name(env.settings.get('network.fail2ban', 'jail.dir'))
	destdir = env.settings.get('network.fail2ban', 'jail.destdir')
	jenable = env.settings.getlist('network.fail2ban', 'jail.enable')
	if env.assets.isdir(jaildir):
		for jn in jenable:
			src = env.assets.name(jaildir, jn + '.conf')
			dst = path.join(destdir, jn + '.conf')
			env.log("%s" % dst)
			builddir.copy(env, src, dst)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque
from unittest.mock import Mock

class MockPath(object):
	_mock = None
	_expect = None
	_return = None
	_default = None
	sep = '/'

	def __init__(self, cfg):
		self._expect = []
		self._return = {}
		self._default = {}
		self._mock = Mock()
		self.isfile = self._mock.isfile
		self.isdir = self._mock.isdir
		self.join = self._mock.join
		self.abspath = self._mock.abspath
		self.normpath = self._mock.normpath
		self.unlink = self._mock.unlink
		self._configure(cfg)

	def _configure(self, cfg):
		self._setDefaults()
		self.isfile.side_effect = self._sideEffect('isfile')
		self.isdir.side_effect = self._sideEffect('isdir')
		self.join.side_effect = self._sideEffect('join')
		self.abspath.side_effect = self._sideEffect('abspath')
		self.normpath.side_effect = self._sideEffect('normpath')
		self.unlink.side_effect = self._sideEffect('unlink')
		if cfg is None:
			return
		self._parseConfig(cfg)

	def _setDefaults(self):
		self._default['isfile'] = True
		self._default['isdir'] = True

	def _parseConfig(self, cfg):
		data = cfg.get('path', fallback = '')
		if data != '':
			for l in data.splitlines():
				l = l.strip()
				if l != '':
					x = l.split(';')
					rtrn = x[0].strip()
					cmdline = ';'.join(x[1:]).strip()
					self._expect.append(cmdline)
					util = cmdline.split(' ')[0].strip()
					self._setReturn(util, rtrn)

	def _setReturn(self, name, data):
		if name == '':
			raise RuntimeError('mock path: util name is empty')
		if self._return.get(name, None) is None:
			self._return[name] = deque()
		self._return[name].appendleft(data)

	def _sideEffect(self, util):
		def wrapper(*args, **kwargs):
			rtrn = self._return.get(util, None)
			if rtrn is None:
				return self._default.get(util, None)
			try:
				data = rtrn.pop()
			except IndexError:
				return self._default.get(util, None)
			if data == '':
				return self._default.get(util, None)
			if data == 'False':
				return False
			elif data == 'True':
				return True
			return data
		return wrapper

	def check(self):
		got = []
		for x in self._mock.mock_calls:
			xname = x[0]
			xargs = x[1]
			cmdline = xname
			if len(xargs) > 0:
				cmdline = "%s %s" % (xname, ' '.join([str(i) for i in xargs]))
			xkwargs = x[2]
			for k, v in xkwargs.items():
				v = str(v)
				cmdline = "%s, %s=%s" % (cmdline, k, v)
			got.append(cmdline)
		assert got == self._expect, \
			"mock path got: %s - expect: %s" % (got, self._expect)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log, version, web
from _sadm.cmd import flags
from _sadm.web import syslog

def _getArgs():
	p = flags.new('sadm-web', desc = 'sadm web interface')
	# ~ p.add_argument('--address', help = 'bind to ip address (localhost)',
		# ~ metavar = 'address', default = 'localhost')
	p.add_argument('--port', help = 'bind to tcp port (3478)',
		metavar = 'number', type = int, default = 3478)
	return flags.parse(p)

def main():
	args = _getArgs()
	log.info("sadm-web v%s" % version.get())
	log.msg("http://%s:%d/" % ('localhost', args.port))
	syslog.init()
	web.start('localhost', args.port, args.debug)
	syslog.close()
	log.msg('done!')

if __name__ == '__main__':
	main()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import path, builddir

__all__ = ['build']

def build(env):
	homedir = env.settings.get('os', 'users.home.dir')
	env.debug("home dir %s" % homedir)
	for user in env.settings['os.user']:
		uid = env.settings.getint('os.user', user)
		# ~ env.log("%d %s" % (uid, user))
		_synchome(env, user, homedir)

def _synchome(env, user, homedir):
	cfgdir = env.settings.get('os', 'users.config.dir')
	dirmode = env.settings.getint('os', 'users.dirmode')
	filemode = env.settings.getint('os', 'users.filemode')
	srcdir = path.join(cfgdir, user)
	dstdir = path.join(homedir, user)
	if env.assets.isdir(cfgdir, user):
		env.log("sync %s -> %s" % (srcdir, dstdir))
		builddir.sync(env, srcdir, dstdir, user = user, group = user,
			dirmode = dirmode, filemode = filemode)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys

# colors

_colored = sys.stdout.isatty() and sys.stderr.isatty()

_colDebug = lambda msg: 'D: ' + msg
_colWarn = lambda msg: 'W: ' + msg
_colError = lambda msg: 'E: ' + msg
_colInfo = lambda msg: 'I: ' + msg
_colMsg = lambda msg: msg

_cyan = '\033[0;36m'
_red = '\033[0;31m'
_yellow = '\033[0;33m'
_blue = '\033[0;34m'
_green = '\033[0;32m'
_grey = '\033[1;30m'
_reset = '\033[0m'

# ~ _colored = False
def _setColored():
	global _colDebug
	global _colWarn
	global _colError
	global _colInfo
	global _colMsg
	if _colored: # pragma: no cover
		_colDebug = lambda msg: _grey + msg + _reset
		_colError = lambda msg: _red + msg + _reset
		_colWarn = lambda msg: _yellow + msg + _reset
		_colInfo = lambda msg: _blue + msg + _reset
		_colMsg = lambda msg: _green + msg + _reset

# debug file info

_idx = __file__.find('_sadm')

def _getCaller(depth):
	inf = sys._getframe(depth)
	return "%s:%d" % (inf.f_code.co_filename[_idx:], inf.f_lineno)

# setup logger

class _sysLogger(object):

	def __init__(self, level, outs = None):
		self._depth = 3
		self._child = _dummyLogger()
		self._outs = sys.stdout
		self._errs = sys.stderr
		if outs is not None:
			self._outs = outs
			self._errs = outs
		self.debug = self._off
		self.warn = self._off
		self.error = self._off
		self.info = self._off
		self.msg = self._off
		self._initLevel(level)

	def _initLevel(self, level):
		# debug: all messages
		if level == 'debug':
			self.debug = self._debug
			self.warn = self._warn
			self.info = self._info
			self.msg = self._msg
			self.error = self._error
		# info: info + warn + msg + error
		elif level == 'info':
			self.debug = self._off
			self.warn = self._warn
			self.info = self._info
			self.msg = self._msg
			self.error = self._error
		# warn: warn + msg + error
		elif level == 'warn':
			self.debug = self._off
			self.info = self._off
			self.warn = self._warn
			self.msg = self._msg
			self.error = self._error
		# msg: msg + error
		elif level == 'msg':
			self.debug = self._off
			self.info = self._off
			self.warn = self._off
			self.msg = self._msg
			self.error = self._error
		# error: only errors
		elif level == 'error':
			self.debug = self._off
			self.warn = self._off
			self.info = self._off
			self.msg = self._off
			self.error = self._error
		# quiet: no messages
		elif level == 'quiet':
			self.debug = self._off
			self.warn = self._off
			self.info = self._off
			self.msg = self._off
			self.error = self._off
		else:
			raise RuntimeError("invalid log level: %s" % level)

	def _off(self, msg, depth = None, tag = None):
		pass

	def _debug(self, msg, depth = None, tag = None):
		if tag is None:
			tag = ''
		else:
			tag = tag.strip()
			tag += ' '
		if depth is None:
			depth = self._depth
		caller = _getCaller(depth)
		print(_colDebug("%s%s: %s" % (tag, caller, msg)), file = self._errs)
		# no debug info for child logger

	def _error(self, msg):
		print(_colError(msg), file = self._errs)
		self._child.error(msg)

	def _warn(self, msg):
		print(_colWarn(msg), file = self._errs)
		self._child.warn(msg)

	def _info(self, msg):
		print(_colInfo(msg), file = self._outs)
		self._child.info(msg)

	def _msg(self, msg):
		print(_colMsg(msg), file = self._outs)
		self._child.msg(msg)

class _dummyLogger(object):

	def __init__(self):
		self._child = None

	def debug(self, msg, depth = None, tag = None):
		pass

	def error(self, msg):
		pass

	def warn(self, msg):
		pass

	def info(self, msg):
		pass

	def msg(self, msg):
		pass

_logger = _dummyLogger()
_curlevel = None

# public methods

def init(level):
	global _logger
	global _curlevel
	_setColored()
	_logger = _sysLogger(level)
	_curlevel = level

def levels():
	return ['debug', 'info', 'warn', 'msg', 'error', 'quiet']

def defaultLevel():
	return 'warn'

def curLevel():
	return _curlevel

def debugEnabled():
	return _curlevel == 'debug'

def debug(msg, depth = None, tag = None):
	_logger.debug(msg, depth = depth, tag = tag)

def error(msg):
	_logger.error(msg)

def warn(msg):
	_logger.warn(msg)

def info(msg):
	_logger.info(msg)

def msg(msg):
	_logger.msg(msg)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import os
import os.path

__all__ = ['sep', 'isfile', 'isdir', 'join', 'abspath', 'normpath', 'unlink']

class _Path(object):
	def __init__(self):
		self.sep = os.path.sep
		self.isfile = os.path.isfile
		self.isdir = os.path.isdir
		self.join = os.path.join
		self.abspath = os.path.abspath
		self.normpath = os.path.normpath
		self.unlink = os.unlink

_path = _Path()

sep = _path.sep

def isfile(name):
	return _path.isfile(name)

def isdir(name):
	return _path.isdir(name)

def join(base, *parts):
	return _path.join(base, *parts)

def abspath(name):
	return _path.abspath(name)

def normpath(name):
	return _path.normpath(name)

def unlink(name):
	_path.unlink(name)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import response, HTTPError, request, HTTP_CODES

from _sadm import log

__all__ = ['init', 'error']

def _handler(code, error):
	log.debug("handler %d" % code)
	log.debug("%d - %s" % (error.status_code, error.status_line))
	argsLen = len(error.args)
	if argsLen >= 3:
		log.error("%s %d - %s" % (request.remote_addr, code, error.args[2]))
		if argsLen >= 4:
			log.debug("%s" % error.args[3])
	else:
		log.error("%s %d - %s" % (request.remote_addr, code, request.path))
	response.headers['Content-Type'] = 'text/plain; charset=UTF-8'
	codeStatus = HTTP_CODES.get(code, None)
	if codeStatus is not None:
		return "%s\n" % codeStatus
	return "ERROR %d\n" % code

def init(wapp):

	@wapp.error(500)
	def error_500(error):
		return _handler(500, error)

	@wapp.error(400)
	def error_400(error):
		return _handler(400, error)

	@wapp.error(404)
	def error_404(error):
		return _handler(404, error)

	@wapp.error(405)
	def error_405(error):
		return _handler(405, error)

def error(code, msg):
	log.error("%s %d - %s" % (request.remote_addr, code, msg))
	return HTTPError(
		status = code,
		body = msg,
	)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from hashlib import sha256, md5
from os import path

class CheckBuilddir(object):
	_env = None
	_rootdir = None
	_deployfn = None
	_envfn = None
	_metadir = None
	_zipfn = None

	def __init__(self, env):
		self._env = "%s/%s" % (env.profile(), env.name())
		self._rootdir = env.build.rootdir()
		self._deployfn = self._rootdir + '.deploy'
		self._envfn = self._rootdir + '.env'
		self._metadir = self._rootdir + '.meta'
		self._zipfn = self._rootdir + '.zip'

	def content(self):
		assert path.isdir(self._rootdir), "%s builddir not found" % self._rootdir
		assert path.isfile(self._deployfn), \
			"%s deploy self-extract file not found" % self._deployfn
		assert path.isfile(self._envfn), \
			"%s deploy .env file not found" % self._envfn
		assert path.isdir(self._metadir), \
			"%s build meta dir not found" % self._metadir
		assert path.isfile(self._zipfn), \
			"%s deploy zip file not found" % self._zipfn

	def envChecksum(self):
		with open(self._envfn, 'r') as fh:
			expect = fh.read().strip().split()[0]
		with open(self._zipfn, 'rb') as fh:
			h = sha256(fh.read())
			got = h.hexdigest()
		assert got == expect, \
			"%s env checksum got: %s - expect: %s" % (self._env, got, expect)

	def file(self, name, content = None, checksum = None):
		fn = path.join(self._rootdir, name)
		assert path.isfile(fn), "%s file not found" % fn
		if content is not None:
			with open(fn, 'r') as fh:
				got = fh.read()
			assert got == content, \
				"%s content got: %s - expect: %s" % (fn, got, content)
		if checksum is not None:
			with open(fn, 'rb') as fh:
				h = md5(fh.read())
				got = h.hexdigest()
			assert got == checksum, \
				"%s checksum got: %s - expect: %s" % (fn, got, checksum)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

try:
	from _sadm._version import version as _version
except ImportError:
	_version = 'master'

try:
	from _sadm._version_build import version_build as _version_build
except ImportError:
	_version_build = 'devel'

__all__ = ['get', 'build', 'string']

def get():
	return _version

def build():
	return _version_build

def string(prog = None):
	s = "%s (build %s)" % (get(), build())
	if prog is None:
		return '%(prog)s version ' + s
	return "%s version %s" % (prog, s)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, unlink
from pytest import raises

from _sadm.env import action as envAction
from _sadm.errors import EnvError, PluginError

def test_run(env_setup):
	env = env_setup()
	cfgfn = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'configure.ini')
	sessfn = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'session.json')
	for fn in (cfgfn, sessfn):
		if path.isfile(fn):
			unlink(fn)
	assert not path.isfile(cfgfn)
	assert not path.isfile(sessfn)
	envAction.run(env, 'build')
	assert path.isfile(cfgfn)
	assert path.isfile(sessfn)

def test_run_invalid_action(testing_env):
	env = testing_env()
	with raises(EnvError, match = 'invalid action configure'):
		envAction.run(env, 'configure')

def test_plugin_no_action(testing_env):
	env = testing_env()
	env.configure()
	with raises(PluginError, match = 'testing plugin no action nocmd'):
		envAction._runAction(env, 'build', cmd = 'nocmd', force = True)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from configparser import ConfigParser
from os import path
from pytest import raises

from _sadm import cfg
from _sadm.errors import ProfileError

def test_cfg():
	assert cfg._DEFAULT['name'] == ''
	assert cfg._DEFAULT['profile'] == 'default'
	assert cfg._DEFAULT['env'] == 'default'
	c = cfg.new()
	assert isinstance(c, ConfigParser)
	assert c.name() == 'sadmtest'
	assert c.filename().endswith(path.join('tdata', 'sadm.cfg'))
	assert len(c.defaults()) == 5
	assert c.get('default', 'name') == 'sadmtest'
	assert c.get('default', 'profile') == 'testing'
	assert c.get('default', 'env') == 'testing'
	assert c.get('default', 'dir') == '.'
	assert c.listPlugins('testing') == ['sadm', 'os', 'testing']
	assert len(c.sections()) == 3
	assert sorted(c.listProfiles()) == ['envsetup', 'plugin', 'testing']
	assert c.has_section('testing')
	assert c.get('testing', 'dir') == './tdata'
	assert c.get('testing', 'env.testing') == 'testing/config.ini'
	assert sorted(c.listEnvs('testing')) == ['testing', 'testing.nodir']

def test_profile_error():
	c = cfg.new()
	with raises(ProfileError, match = 'config profile noprofile not found'):
		c.listEnvs('noprofile')

def test_default_plugins():
	assert tuple(sorted(cfg._enablePlugins)) == ('sadm', 'sadmenv')

def test_name():
	c = cfg.new()
	assert c.name() == 'sadmtest'
	assert c._getName() == 'tdata'

def test_file_not_found():
	fn = path.join('tdata', 'testing', 'noconfig.ini')
	with raises(ProfileError, match = 'noconfig.ini file not found'):
		c = cfg.new(cfgfile = fn)

def test_noname():
	fn = path.join('tdata', 'testing', 'config-noname.ini')
	c = cfg.new(cfgfile = fn)
	assert c.name() == 'testing'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import bottle
from configparser import ConfigParser, ExtendedInterpolation

from _sadm import libdir, log, version
from _sadm.listen import errors
from _sadm.utils import path

__all__ = ['wapp', 'config', 'init']

wapp = bottle.Bottle()
bottle.TEMPLATE_PATH = []

_cfgfn = libdir.fpath('listen', 'wapp.conf')
wapp.config.load_config(_cfgfn)

config = ConfigParser(
	defaults = {
		'sadm': {
			'log': 'warn',
		},
		'sadm.listen': {
			'host': '127.0.0.1',
			'port': 3666,
		}
	},
	allow_no_value = False,
	delimiters = ('=',),
	comment_prefixes = ('#',),
	strict = True,
	interpolation = ExtendedInterpolation(),
	default_section = 'default',
)

def init(cfgfn = None):
	if cfgfn is None:
		cfgfn = path.join(path.sep, 'etc', 'opt', 'sadm', 'listen.cfg')
	config.read(cfgfn)

	log.init(config.get('sadm', 'log', fallback = 'warn'))
	log.debug(version.string('sadm'))

	errors.init(wapp)

	from _sadm.listen import handlers
	from _sadm.listen.plugin import HandlersPlugin
	wapp.install(HandlersPlugin())

	initDone = {}
	for sect in config.sections():

		if sect.startswith('sadm.webhook:'):
			if not initDone.get('webhook', False):
				log.debug('enable webhook')
				from _sadm.listen import webhook
				from _sadm.listen.plugin.webhook import WebhookPlugin
				wapp.install(WebhookPlugin())
				initDone['webhook'] = True

	return wapp



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import argparse

from _sadm import log, version

__all__ = ['new', 'parse']

def new(prog, desc = ''):
	p = argparse.ArgumentParser(prog = prog, description = desc)
	p.add_argument('-V', '--version', help = 'show version and exit',
		action = 'version', version = version.string())
	p.add_argument('--debug', help = 'enable debug settings',
		action = 'store_true', default = False)
	p.add_argument('--log', help = "set log level (%s)" % log.defaultLevel(),
		default = log.defaultLevel(), choices = log.levels())
	p.add_argument('--env', help = 'env name (default)',
		metavar = 'name', default = 'default')
	p.add_argument('--profile', help = 'profile name (default)',
		metavar = 'name', default = 'default')
	return p

def parse(p):
	args = p.parse_args()
	if args.debug:
		log.init('debug')
	else:
		log.init(args.log)
	log.debug("sadm version %s" % version.get())
	return args



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import tarfile

from collections import deque, namedtuple
from os import path

from _sadm import asset
from _sadm.errors import BuildError

_Info = namedtuple('_Info', ('name', 'args', 'type'))

class Manager(asset.Manager):
	_tarfn = None
	_data = None

	def create(self):
		self._data = deque()
		rdir = self.rootdir()
		en = path.basename(rdir)
		mdir = path.normpath(rdir) + '.meta'
		self._tarfn = path.join(mdir, en + '.tar')
		if path.isfile(self._tarfn):
			raise BuildError("%s file exists" % self._tarfn)

	def close(self):
		with tarfile.open(self._tarfn, 'w') as tar:
			for inf in self._data:
				if inf.type == 'dir':
					self._adddir(tar, inf)
					continue
				self._addfile(tar, inf)

	def _tarinfo(self, inf, user = 'root', group = '',
		filemode = 644, dirmode = 755):
		if group == '':
			group = user
		inf.mtime = 0 # reproducible builds
		inf.uid = 0
		inf.gid = 0
		inf.uname = user
		inf.gname = group
		if inf.isdir():
			inf.mode = int(str(dirmode), 8)
		else:
			inf.mode = int(str(filemode), 8)

	def addfile(self, name, **kwargs):
		self._data.append(_Info(name = name, args = kwargs, type = 'file'))

	def _addfile(self, tar, inf):
		def _filter(fi):
			self._tarinfo(fi, **inf.args)
			return fi
		arcname = self.name(inf.name)
		fn = path.join(self.rootdir(), arcname)
		tar.add(fn, arcname = arcname, recursive = False, filter = _filter)

	def adddir(self, name, **kwargs):
		self._data.append(_Info(name = name, args = kwargs, type = 'dir'))

	def _adddir(self, tar, inf):
		def _filter(fi):
			self._tarinfo(fi, **inf.args)
			return fi
		arcname = self.name(inf.name)
		dirpath = path.join(self.rootdir(), arcname)
		tar.add(dirpath, arcname = arcname, recursive = True, filter = _filter)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path
from pytest import raises

from _sadm.errors import PluginError, PluginScriptNotFound, PluginScriptNoExec
from _sadm.errors import PluginScriptTimeout
from _sadm.utils import scripts

sdir = path.join(path.dirname(path.dirname(path.dirname(__file__))), 'tlib',
	'_sadmtest', 'scripts', 'testing', 'testing')
s = scripts.Scripts('testing', 'testing', sdir = sdir)

def test_scripts_dir():
	assert s._dir == sdir

def test_noscript():
	spath = path.join(sdir, 'noscript.sh')
	with raises(PluginScriptNotFound, match = "PluginScriptNotFound: %s" % spath) as err:
		s.run('noscript.sh')
	assert err.errisinstance(PluginError)

def test_run():
	rc = s.run('testing.sh')
	assert rc == 0

def test_run_error():
	rc = s.run('testing-error.sh')
	assert rc == 1

def test_run_noexec():
	spath = path.join(sdir, 'testing-noexec.sh')
	with raises(PluginScriptNoExec, match = "PluginScriptNoExec: %s" % spath) as err:
		s.run('testing-noexec.sh')
	assert err.errisinstance(PluginError)

def test_timeout():
	spath = path.join(sdir, 'testing-timeout.sh')
	prevttl = scripts._TTL
	scripts._TTL = 0.2
	with raises(PluginScriptTimeout,
		match = "PluginScriptTimeout: %s after 0.2 seconds" % spath) as err:
		s.run('testing-timeout.sh', '1')
	assert err.errisinstance(PluginError)
	scripts._TTL = prevttl

def test_run_args():
	rc = s.run('testing-args.sh', ('1', '2', '3'))
	assert rc == 0



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm.configure import pluginsList
from _sadm.env.settings import Settings

_srcdir = path.dirname(path.dirname(path.dirname(__file__)))

def test_build_testing(testing_plugin):
	print('-- build plugin: testing')
	p = testing_plugin('testing', ns = '_sadmtest', cfgfn = 'build.ini')
	p.build()
	p.check.builddir.file('sadm.testing') # just to check it works both ways
	p.check.builddir.file('sadm.testing', 'testing\n')

# ~ def test_build_vcs_clone(testing_plugin):
	# ~ print('-- build plugin: vcs.clone')
	# ~ p = testing_plugin('vcs.clone', cfgfn = 'build.ini')
	# ~ p.build()

def test_all_build(testing_plugin):
	pl = list(pluginsList())
	print('-- plugins list:', pl)
	for n in pl:
		if n == 'testing':
			continue
		print('-- build plugin:', n)
		_pbuild(testing_plugin, n, 'build.ini')

def _pbuild(pnew, pname, cfgfn):
	fn = path.join(_srcdir, 'tdata', 'plugin',
		pname.replace('.', path.sep), 'config', cfgfn)
	if path.isfile(fn):
		p = pnew(pname, cfgfn = cfgfn)
		p.build()
		# check assets/testing.txt was sync'ed
		# so, all build tests should do that, yes...
		p.check.builddir.file('testing.txt', 'testing\n')
		_pcheck(p, fn)

def _pcheck(p, fn):
	pcfg = Settings()
	with open(fn, 'r') as fh:
		pcfg.read_file(fh)
	if pcfg.has_section('_sadmtest.check'):
		for opt, val in pcfg.items('_sadmtest.check'):
			if opt.startswith('builddir.file.'):
				_checkBuilddirFile(p, val)

def _checkBuilddirFile(p, args):
	fn = args.split()[0]
	if fn.startswith(path.sep):
		fn = fn.replace(path.sep, '', 1)
	try:
		cksum = args.split()[1]
	except IndexError:
		cksum = None
	p.check.builddir.file(fn, checksum = cksum)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys

from hashlib import sha256 as digest
from io import StringIO
from os import path

from _sadm.configure import getPlugin

from _sadmtest import mock
from _sadmtest.check import CheckEnv

__all__ = ['Plugin']

_srcdir = path.dirname(path.dirname(path.dirname(path.dirname(__file__))))

class Plugin(object):
	_p = None
	_env = None
	_envAction = None
	check = None

	def __init__(self, name, env, ns = '_sadm'):
		self._p = getPlugin(name, 'configure')
		assert self._p.name == name, \
			"plugin %s name error: %s" % (name, self._p.name)
		assert self._p.fullname == "%s.plugin.%s" % (ns, name), \
			"plugin %s fullname error: %s" % (name, self._p.fullname)
		cfgfn = path.join(ns, 'plugin', name.replace('.', path.sep), 'config.ini')
		assert self._p.config.endswith(cfgfn), \
			"plugin %s config error: got: %s - expect: %s" % (name, self._p.config, cfgfn)
		assert self._p.meta.endswith(path.join(ns,
			'plugin', name.replace('.', path.sep), 'meta.json')), \
			"plugin %s meta file error: %s" % (name, self._p.meta)
		self._env = env
		assert self._envSettings() == '', \
			"env %s settings are not empty: %s" % (env.name(), self._envSettings())
		self.check = CheckEnv(self._env)
		from _sadm.env import action as envAction
		self._envAction = envAction

	def _error(self, *args):
		print('ERROR:', *args, file = sys.stderr)
		return False

	def _envSettings(self):
		buf = StringIO()
		self._env.settings.write(buf)
		buf.seek(0, 0)
		return buf.read()

	def _cksum(self, data):
		h = digest(data.encode('utf-8'))
		return h.hexdigest()

	def configure(self):
		self._env.configure()
		fn = path.join(_srcdir, 'tdata', 'plugin',
			self._p.name.replace('.', path.sep), 'default.ini')
		if path.isfile(fn):
			with open(fn, 'r') as fh:
				expect = self._cksum(fh.read())
			got = self._cksum(self._envSettings())
			if got != expect:
				self._error('--')
				self._error(fn)
				self._error('--')
				self._error('expect:', expect)
				with open(fn, 'r') as fh:
					self._error(fh.read())
				self._error('--')
				self._error('   got:', got)
				self._error(self._envSettings())
				self._error('--')
				return False
		else:
			return self._error(fn, 'file not found')
		return True

	def build(self):
		self._envAction.run(self._env, 'build')
		self.check.builddir.content()
		self.check.builddir.envChecksum()

	def deploy(self, mockCfg = 'build.ini', mockProfile = None):
		from _sadm.env.settings import Settings
		cfgfn = path.join('tdata', 'plugin',
			self._p.name.replace('.', path.sep), 'config', mockCfg)
		cfg = Settings()
		with open(cfgfn, 'r') as fh:
			cfg.read_file(fh)
		cfgSection = '_sadmtest.mock.deploy'
		if mockProfile is not None:
			cfgSection = "%s.%s" % (cfgSection, mockProfile)
		if cfg.has_section(cfgSection):
			cfg = cfg[cfgSection]
		else:
			cfg = None
		with mock.deploy(self._p.name, cfg):
			self._envAction.run(self._env, 'deploy')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import request

from _sadm.listen.wapp import wapp, config

from .repo import WebhookRepo

__all__ = ['repo']

@wapp.route('/hook/<provider>/<name>/<action>', 'POST')
def repo(provider, name, action):
	repo = WebhookRepo(config, provider, name)
	repo.auth(request)
	repo.exec(request, action)
	return 'OK\n'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, unlink
from pytest import raises

from _sadm.errors import SessionError, EnvError
from _sadm.utils import builddir

def test_lock_unlock(env_setup):
	lockfn = path.join('tdata', 'build', 'envsetup', 'testing.lock')
	env = env_setup(configure = True)
	assert not path.isfile(lockfn)
	builddir.lock(env)
	assert path.isfile(lockfn)
	builddir.unlock(env)
	assert not path.isfile(lockfn)

def test_lock_error(env_setup):
	lockfn = path.join('tdata', 'build', 'envsetup', 'testing.lock')
	env = env_setup(configure = True)
	assert not path.isfile(lockfn)
	builddir.lock(env)
	with raises(SessionError, match = 'lockfn option already set'):
		builddir.lock(env)
	del env.session._d['lockfn']
	with raises(EnvError, match = 'lock file exists: '):
		builddir.lock(env)
	builddir.unlock(env)

def test_unlock_error(env_setup):
	lockfn = path.join('tdata', 'build', 'envsetup', 'testing.lock')
	if path.isfile(lockfn):
		unlink(lockfn)
	env = env_setup(configure = True)
	assert not path.isfile(lockfn)
	with raises(SessionError, match = 'lockfn option not set'):
		builddir.unlock(env)
	builddir.lock(env) # set session lockfn
	builddir.unlock(env)
	with raises(EnvError, match = 'unlock file not found'):
		builddir.unlock(env)

def test_fpath(env_setup):
	env = env_setup(configure = True)
	builddir.lock(env)
	try:
		p = path.join('tdata', 'build', 'envsetup', 'testing', 'p1', 'p2', 'p3')
		assert builddir.fpath(env, 'p1', 'p2', 'p3') == path.realpath(p)
	finally:
		builddir.unlock(env)

def test_fpath_relname(env_setup):
	env = env_setup(configure = True)
	builddir.lock(env)
	try:
		p = path.join('tdata', 'build', 'envsetup', 'testing', 'p1', 'p2', 'p3')
		assert builddir.fpath(env, path.sep + 'p1', 'p2', 'p3') == path.realpath(p)
	finally:
		builddir.unlock(env)

def test_fpath_meta(env_setup):
	env = env_setup(configure = True)
	builddir.lock(env)
	try:
		p = path.join('tdata', 'build', 'envsetup', 'testing.meta', 'p1')
		assert builddir.fpath(env, 'p1', meta = True) == path.realpath(p)
	finally:
		builddir.unlock(env)

def test_create(env_setup):
	env = env_setup(configure = True)
	builddir.lock(env)
	try:
		env.build.create()
		fn = path.join('tdata', 'build', 'envsetup', 'testing', 'file.test')
		assert not path.isfile(fn)
		with builddir.create(env, 'file.test') as fh:
			fh.write('1')
		assert path.isfile(fn)
	finally:
		env.build.close()
		builddir.unlock(env)

def test_cleandir(env_setup):
	env = env_setup(configure = True)
	builddir.lock(env)
	try:
		envfn = path.join('tdata', 'build', 'envsetup', 'testing.env')
		metadir = path.join('tdata', 'build', 'envsetup', 'testing.meta')
		assert path.isdir(metadir)
		assert not path.isfile(envfn)
		with open(envfn, 'x') as fh:
			fh.write('1')
		assert path.isfile(envfn)
		builddir._cleandir(env, env.build.rootdir())
		assert path.isdir(metadir)
		assert not path.isfile(envfn)
	finally:
		builddir.unlock(env)

def test_create_makedirs(env_setup):
	fn = path.join('asset', 'create.test')
	env = env_setup(name = 'build.test', action = 'build')
	assert not path.isdir(env.build.rootdir('asset'))
	with builddir.create(env, fn) as fh:
		fh.write('1')
	assert path.isfile(env.build.rootdir(fn))

def test_sync_error(env_setup):
	srcdir = 'nosrcdir'
	dstdir = path.join(path.sep, 'internal', 'path')
	env = env_setup(name = 'build.test', action = 'build')
	assert not path.isdir(env.build.rootdir('internal'))
	with raises(FileNotFoundError, match = "No such file or directory"):
		builddir.sync(env, srcdir, dstdir)

def test_sync(env_setup):
	srcdir = path.join('build.test', 'assets.sync')
	dstdir = path.join(path.sep, 'internal', 'path')
	env = env_setup(name = 'build.test', action = 'build')
	assert not path.isdir(env.build.rootdir('internal'))
	builddir.sync(env, srcdir, dstdir)
	assert path.isfile(env.build.rootdir(dstdir, 'file.test'))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json

from _sadm import log
from _sadm.errors import CommandError
from _sadm.listen.exec import dispatch
from _sadm.listen.errors import error

from .provider.bitbucket import BitbucketProvider

__all__ = ['WebhookRepo']

_validVCS = {
	'git': True,
}
_provider = {
	'bitbucket': BitbucketProvider(),
}

class WebhookRepo(object):
	_cfg = None
	_prov = None
	_provName = None
	_repoName = None
	_repoVCS = None

	def __init__(self, config, provider, name):
		self._provName = provider
		self._repoName = name
		prov = _provider.get(provider, None)
		if prov is None:
			raise error(400, "webhook invalid provider: %s" % provider)
		sect = "sadm.webhook:%s" % name
		if not config.has_section(sect):
			raise error(400, "webhook %s repo not found: %s" % (provider, name))
		self._cfg = config[sect]

		self._prov = prov
		self._loadProvider(self._cfg, provider, name)
		self._checkRepo(self._cfg)

	def _loadProvider(self, cfg, provider, name):
		rProv = cfg.get('provider', fallback = 'none')
		if rProv != provider:
			raise error(400, "webhook %s repo %s invalid provider: %s" % (provider, name, rProv))

	def _checkRepo(self, cfg):
		vcs = cfg.get('vcs', fallback = 'git')
		if not _validVCS.get(vcs, False):
			raise error(400, "webhook %s repo %s invalid vcs: %s" % (provider, name, vcs))
		self._repoVCS = vcs
		# TODO: check repo.path and other pre-fly checks

	def auth(self, req):
		self._prov.auth(req, self._cfg)

	def exec(self, req, action):
		# TODO: check self._cfg.getboolean(action... if disabled raise error 400
		log.debug("req.body: %s" % req.body)
		if not req.body:
			raise error(400, "webhook %s repo %s no request body" % (self._provName, self._repoName))
		obj = json.loads(req.body.read())
		args = self._prov.repoArgs(obj, self._cfg)
		task = "webhook.repo.%s" % self._repoVCS
		try:
			dispatch(req, task, action, args)
		except CommandError as err:
			raise error(500, "webhook %s repo %s: %s" % (self._provName, self._repoName, err))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import shutil

__all__ = ['configure']

def configure(env, cfg):
	env.debug("archive formats: %s" % [a[0] for a in shutil.get_archive_formats()])
	env.settings.merge(cfg, 'sadmenv', ('target.dir',))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from configparser import ConfigParser, ExtendedInterpolation
from configparser import Error as ParserError

from _sadm import log
from _sadm.errors import ServiceError

class Service(ConfigParser):

	def __init__(self, cfgfn):
		super().__init__(defaults = {},
			allow_no_value = False,
			delimiters = ('=',),
			comment_prefixes = ('#',),
			strict = True,
			interpolation = ExtendedInterpolation(),
			default_section = 'default')
		self.__read(cfgfn)

	def __read(self, fn):
		log.debug("read file %s" % fn)
		try:
			super().read(fn)
		except ParserError as err:
			raise ServiceError(str(err))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import path, systemd

__all__ = ['deploy']

def deploy(env):
	destdir = env.settings.get('network.fail2ban', 'jail.destdir')
	jdisable = env.settings.getlist('network.fail2ban', 'jail.disable')
	for jn in jdisable:
		fn = path.join(destdir, jn + '.conf')
		if path.isfile(fn):
			env.log("remove %s" % fn)
			path.unlink(fn)
	systemd.restart('fail2ban')
	systemd.status('fail2ban')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from time import strftime, time

from _sadm.errors import PluginError

__all__ = ['run']

_validAction = {
	'build': True,
	'deploy': True,
}

def run(env, action):
	_start = time()
	env.info("%s start %s" % (action, strftime('%c %z')))
	env.log("%s %s" % (env.config.name(), env.config.filename()))
	try:
		if not _validAction.get(action, False):
			raise env.error("invalid action %s" % action)
		with env.lock() as env:
			env.configure()
			_runPreAction(env, action)
			_runAction(env, action)
			_runPostAction(env, action)
			env.report(action, startTime = _start)
	finally:
		env.info("%s end %s" % (action, strftime('%c %z')))

def _runAction(env, action, cmd = None, force = False, revert = False):
	if cmd is None:
		cmd = action
	for p in env.plugins(action, revert = revert):
		if hasattr(p.mod, cmd):
			func = getattr(p.mod, cmd)
			tag = "%s.%s" % (cmd, p.name)
			env.start(tag)
			func(env)
			env.end(tag)
		else:
			env.debug("%s plugin no action %s" % (p.name, cmd))
			if force:
				raise PluginError("%s plugin no action %s" % (p.name, cmd))

def _runPreAction(env, action):
	_runAction(env, action, cmd = "pre_%s" % action)

def _runPostAction(env, action):
	_runAction(env, action, cmd = "post_%s" % action, revert = True)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import cmd
from _sadm.utils.scripts import Scripts

__all__ = ['deploy']

def deploy(env):
	scripts = Scripts('service.apache', env.dist())
	_reload = False
	args = env.settings.getlist('service.apache', 'conf.disable')
	if len(args) > 0:
		env.log('conf.disable')
		_reload = True
		scripts.run('disconf.sh', args)
	args = env.settings.getlist('service.apache', 'mod.disable')
	if len(args) > 0:
		env.log('mod.disable')
		_reload = True
		scripts.run('dismod.sh', args)
	args = env.settings.getlist('service.apache', 'site.disable')
	if len(args) > 0:
		env.log('site.disable')
		_reload = True
		scripts.run('dissite.sh', args)
	args = env.settings.getlist('service.apache', 'conf.enable')
	if len(args) > 0:
		env.log('conf.enable')
		_reload = True
		scripts.run('enconf.sh', args)
	args = env.settings.getlist('service.apache', 'mod.enable')
	if len(args) > 0:
		env.log('mod.enable')
		_reload = True
		scripts.run('enmod.sh', args)
	args = env.settings.getlist('service.apache', 'site.enable')
	if len(args) > 0:
		env.log('site.enable')
		_reload = True
		scripts.run('ensite.sh', args)
	if _reload:
		scripts.run('reload.sh')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path, makedirs

def test_deploy_testing(testing_plugin):
	makedirs(path.join('tdata', 'deploy', 'plugin'), exist_ok = True)
	p = testing_plugin('testing', ns = '_sadmtest', deploy = True)
	print('-- deploy plugin: testing')
	p.deploy()

def test_all_deploy(testing_plugin):
	makedirs(path.join('tdata', 'deploy', 'plugin'), exist_ok = True)
	t = testing_plugin(ns = '_sadmtest', deploy = True, buildDeploy = False)
	for opt in t._env.config.options('deploy'):
		if opt.startswith('env.'):
			pname = '.'.join(opt.split('.')[1:])
			if pname == 'testing':
				continue
			p = testing_plugin(pname, deploy = True)
			print('-- deploy plugin:', pname)
			p.deploy()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import systemd
from _sadm.utils.sh import makedirs, chmod, chown

__all__ = ['deploy']

def deploy(env):
	dbdir = env.settings.get('service.munin', 'db.dir')
	makedirs(dbdir, exists_ok = True)
	chmod(dbdir, 0o750)

	dbuser = env.settings.get('service.munin', 'dbdir.user')
	dbgroup = env.settings.get('service.munin', 'dbdir.group')
	chown(dbdir, user = dbuser, group = dbgroup)

	env.log("dbdir %s (%s:%s)" % (dbdir, dbuser, dbgroup))

	if systemd.status('cron') == 0:
		systemd.reload('cron')
	else:
		systemd.start('cron')

	systemd.start('munin')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log, cfg
from _sadm.errors import ProfileError

__all__ = ['Profile']

class Profile(object):
	_name = None

	def __init__(self, name, config = None):
		log.debug("profile init %s" % name)
		if config is None:
			config = cfg.new()
		self._name = name
		self._load(config)

	def _load(self, config):
		log.debug("load %s" % self._name)
		if not config.has_section(self._name):
			log.debug("%s profile not found" % self._name)
			raise ProfileError("%s profile not found" % self._name)

	def name(self):
		return self._name



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json
import ssl

from urllib.request import urlopen

from _sadm import log

__all__ = ['ListenClient']

class ListenClient(object):
	_url = None

	def __init__(self, url):
		log.debug("url %s" % url)
		self._url = url

	def _path(self, *parts):
		return '/'.join(parts)

	def _post(self, path, data):
		url = self._path(self._url, path)
		ctx = None
		if url.startswith('https'):
			ctx = ssl.create_default_context()
			ctx.check_hostname = False
		try:
			with urlopen(url, data, context = ctx) as resp:
				if resp.status != 200:
					log.error("%s returned: %d - %s" % (url, resp.status, resp.reason))
		except Exception as err:
			log.error("%s - %s" % (url, err))

	def exec(self, task, action, args):
		log.debug("exec: %s %s" % (task, action))
		path = self._path('_', 'exec', task, action)
		self._post(path, json.dumps(args).encode('UTF-8'))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm import log
from _sadm.deploy import cmd

def cmdArgs(parser):
	p = parser.add_parser('deploy', help = 'deploy sadm.env')
	p.set_defaults(command = 'deploy')

def main(args):
	log.debug("deploy %s" % args.env)
	return cmd.run(args.env)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from .builddir import CheckBuilddir

class CheckEnv(object):
	_env = None
	builddir = None

	def __init__(self, env):
		self._env = env
		self.builddir = CheckBuilddir(self._env)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import version
from _sadm.utils import path

def configure(env, cfg):
	s = env.settings
	s.set('sadm', 'env', env.name())
	s.set('sadm', 'profile', env.profile())
	sess = env.session
	sess.set('sadm.version', version.get())
	_configureDeploy(env, cfg)
	_configureListen(env, cfg)

def _configureDeploy(env, cfg):
	if env.assets.isfile('deploy.cfg'):
		env.log('sync deploy.cfg')
		if not cfg.has_section('sync'):
			cfg.add_section('sync')
		cfg.set('sync', 'sadm.deploy.config',
			"deploy.cfg %s filemode=644" % path.join(path.sep,
				'etc', 'opt', 'sadm', 'deploy.cfg'))

def _configureListen(env, cfg):
	if env.assets.isfile('listen.cfg'):
		env.log('enable sadm.listen')
		env.session.set('sadm.listen.enable', True)
		# sync
		if not cfg.has_section('sync'):
			cfg.add_section('sync')
		cfg.set('sync', 'sadm.listen.config',
			"listen.cfg %s filemode=644" % path.join(path.sep,
				'etc', 'opt', 'sadm', 'listen.cfg'))
		# os.pkg
		if not env.settings.has_section('os.pkg'):
			env.settings.add_section('os.pkg')
		if env.dist() == 'debian':
			env.settings.setlist('os.pkg', 'debian.sadm.listen.install', (
				'at', 'uwsgi-plugin-python3', 'ssl-cert',
			))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from pytest import raises
from _sadm.errors import Error
from _sadm.env.profile import Profile

def test_profile(testing_profile):
	p = testing_profile()
	assert p.name() == 'testing'

def test_profileError():
	with raises(Error, match = 'ProfileError: noname profile not found'):
		Profile('noname')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
import sqlite3
from os import path, makedirs

from _sadm import log

__all__ = ['init', 'close', 'last']

# SQL stuff

_DB_CREATE = """
CREATE TABLE syslog (
	id INTEGER PRIMARY KEY AUTOINCREMENT,
	time TEXT DEFAULT CURRENT_TIMESTAMP,
	level INTEGER,
	msg TEXT
);
"""
_LVLMAP = {
	'error': 3,
	'warn': 2,
	'info': 1,
	'msg': 0,
}
_LOG_INSERT = 'INSERT INTO syslog (level, msg) VALUES (?, ?)'
_GET_MSG_ID = 'SELECT id, time, level, msg FROM syslog WHERE id = ?'
_GET_LAST = 'SELECT id, time, level, msg FROM syslog ORDER BY id DESC LIMIT ?'

# underlying logger class

class _webLogger(object):
	db = None

	def error(self, msg):
		self.db.execute(_LOG_INSERT, (_LVLMAP['error'], msg))
		self.db.commit()
		# ~ print("[syslog] E:", msg, file = sys.stderr)

	def warn(self, msg):
		self.db.execute(_LOG_INSERT, (_LVLMAP['warn'], msg))
		self.db.commit()
		# ~ print("[syslog] W:", msg, file = sys.stderr)

	def info(self, msg):
		self.db.execute(_LOG_INSERT, (_LVLMAP['info'], msg))
		self.db.commit()
		# ~ print("[syslog] I:", msg, file = sys.stdout)

	def msg(self, msg):
		self.db.execute(_LOG_INSERT, (_LVLMAP['msg'], msg))
		self.db.commit()
		# ~ print("[syslog]", msg, file = sys.stdout)

def _dbInit():
	log.debug("syslog init %s" % _dbfile)
	log.debug("sqlite version %s (lib %s)" % (sqlite3.version, sqlite3.sqlite_version))
	mkdb = not path.isfile(_dbfile)
	if mkdb:
		dbdir = path.dirname(_dbfile)
		makedirs(dbdir, mode = 0o700, exist_ok = True)
	db = sqlite3.connect(_dbfile)
	db.row_factory = sqlite3.Row
	return _dbCheck(db, mkdb)

def _dbCheck(db, create):
	if create:
		log.debug('create syslog db')
		db.executescript(_DB_CREATE)
		db.commit()
	return db

# default logger

_logger = _webLogger()
_dbfile = path.expanduser('~/.local/sadm/syslog.db')

def _getMsgId(msgid):
	return _logger.db.execute(_GET_MSG_ID, [str(msgid)]).fetchone()

# public methods

def init():
	log.debug('init web syslog')
	try:
		_logger.db = _dbInit()
	except Exception as err:
		log.debug("exception: %s" % type(err))
		log.error("could not init web syslog: %s" % err)
	else:
		log.debug('attach child logger')
		log._logger._child = _logger
		log.info('syslog start')

def close():
	if _logger.db is not None:
		log.info('syslog end')
		_logger.db.close()
		log._logger._child = log._dummyLogger()

def last(limit):
	msgs = _logger.db.execute(_GET_LAST, [str(limit)]).fetchall()
	msgs.reverse()
	return msgs



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['configure']

def configure(env, cfg):
	env.settings.merge(cfg, 'service.apache', (
		'sites.dir',
		'conf.disable',
		'mod.disable',
		'site.disable',
		'conf.enable',
		'mod.enable',
		'site.enable',
	))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.web import tpl

def test_loaded():
	assert tpl._viewreg == {
		'home': True,
		'profile': True,
		'syslog': True,
		'about': True,
		'err500': True,
		'err400': True,
	}, 'missing view'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils.cmd import callCheck
from _sadm.utils.sh import chdir, getcwd

from .check import check

__all__ = ['deploy']

def deploy(env):
	status = check(env)
	for st, name, typ, repo in status:
		if st == 'MISS':
			_cloneRepo(env, name, typ, repo)
		else:
			_updateRepo(env, name, typ, repo)

def _cloneRepo(env, name, typ, repo):
	if typ == 'git':
		_gitClone(env, name, repo)

def _updateRepo(env, name, typ, repo):
	if not repo['update']:
		return
	if typ == 'git':
		_gitUpdate(env, name, repo)

_gitConfigDone = False

def _gitConfig():
	global _gitConfigDone
	cmd = ['git', 'config', '--global', 'core.sshCommand',
		'ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no']
	callCheck(cmd)
	_gitConfigDone = True

def _gitClone(env, name, repo):
	_gitConfigDone or _gitConfig()
	env.log("clone git repo %s" % name)
	cmd = ['git', 'clone', '-b', repo['branch'], repo['remote'], repo['path']]
	callCheck(cmd)
	if repo['checkout'] != '':
		oldwd = getcwd()
		try:
			chdir(repo['path'])
			cmd = ['git', 'checkout', repo['checkout']]
			callCheck(cmd)
		finally:
			chdir(oldwd)

def _gitUpdate(env, name, repo):
	env.log("update git repo %s" % name)
	oldwd = getcwd()
	try:
		chdir(repo['path'])
		callCheck(['git', 'pull'])
	finally:
		chdir(oldwd)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import getenv
from io import StringIO

from _sadm import log, cfg
from _sadm.web import tpl
from _sadm.web.app import wapp, view

@wapp.route('/')
@view('index.html')
@tpl.data('home')
def index():
	log.debug("index")
	config = cfg.new()
	return {
		'cfgfile': config.filename(),
		'cfg': _getCfg(config),
		'user': getenv('USER', 'nouser'),
	}

def _getCfg(config):
	buf = StringIO()
	config.reload()
	config.write(buf)
	buf.seek(0, 0)
	return buf.read()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque
from os import path
from unittest.mock import Mock, call

from _sadm import libdir

_srcdir = path.dirname(libdir.root())

class MockCmdProc(object):
	_mock = None
	_cfg = None
	_expect = None
	call = None
	check_call = None

	def __init__(self, cfg):
		self._mock = Mock()
		self.call = self._mock.mock_call
		self.check_call = self._mock.mock_check_call
		self._configure(cfg)

	def _configure(self, cfg):
		self._cfg = self._parseConfig(cfg)
		self.call.side_effect = self._sideEffect('call')
		self.check_call.side_effect = self._sideEffect('check_call')

	def _parseConfig(self, cfg):
		d = {}
		if cfg is None:
			return d
		self._expect = {
			'call': deque(),
			'check_call': deque(),
		}
		d['call'] = self._parseCmdOptions(cfg, 'call')
		d['check_call'] = self._parseCmdOptions(cfg, 'check_call')
		return d

	def _parseCmdOptions(self, cfg, opt):
		d = {}
		data = cfg.get('cmd.' + opt, fallback = '')
		if data != '':
			for line in data.splitlines():
				line = line.strip()
				if line == '':
					continue
				elif line.startswith('#'):
					continue
				i = line.split(';')
				retval = int(i[0])
				cmdline = ';'.join(i[1:]).strip()
				if cmdline.startswith(path.join('_sadm', 'scripts')):
					cmdline = "%s%s%s" % (_srcdir, path.sep, cmdline)
				d[cmdline] = retval
				self._expect[opt].append(cmdline)
		return d

	def _sideEffect(self, method):
		def wrapper(args, **kwargs):
			if not method in self._cfg:
				raise KeyError("%s method is not configured (yet)" % method)
			if isinstance(args, list):
				cmdline = ' '.join(args)
			else:
				cmdline = args
			try:
				return self._cfg[method][cmdline]
			except KeyError as err:
				raise KeyError("%s - %s method" % (str(err), method))
		return wrapper

	def check(self):
		if self._expect is None:
			return
		# call
		got = []
		for x in [x[1][0] for x in self.call.mock_calls]:
			if isinstance(x, list):
				x = ' '.join(x)
			got.append(x)
		expect = list(self._expect['call'])
		assert got == expect, \
			"mock cmd call got: %s - expect: %s" % (got, expect)
		# check_call
		got = []
		for x in [x[1][0] for x in self.check_call.mock_calls]:
			if isinstance(x, list):
				x = ' '.join(x)
			got.append(x)
		expect = list(self._expect['check_call'])
		assert got == expect, \
			"mock cmd check_call \ngot:\n%s\nexpect:\n%s" % ('\n'.join(got), '\n'.join(expect))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import unlink, path
from sqlite3 import Connection, Row

from _sadm import log
from _sadm.web import syslog

_dbfile = './tdata/tmp/web/syslog.db'
syslog._dbfile = _dbfile

def _cleanup():
	if path.isfile(_dbfile):
		unlink(_dbfile)
	assert not path.isfile(_dbfile), 'could not cleanup'

def _dbInit():
	assert syslog._dbfile == _dbfile, 'not using testing db file'
	assert not path.isfile(_dbfile), 'dbfile already exists!'
	return syslog._dbInit()

_cleanup()

def test_dbInit():
	db = _dbInit()
	assert isinstance(db, Connection), 'wrong instance'
	assert db.row_factory == Row, 'wrong db row_factory'
	db.close()
	assert path.isfile(_dbfile), 'db not created'
	_cleanup()

def test_initClose():
	assert isinstance(syslog._logger, syslog._webLogger), \
		'wrong instance'
	assert syslog._logger.db is None, \
		'_logger.db should be None'
	assert isinstance(log._logger, log._dummyLogger), \
		'main logger wrong instance'
	assert log._logger._child is None, \
		'main logger._child should be None'
	syslog.init()
	assert isinstance(syslog._logger.db, Connection), \
		'wrong db instance'
	assert isinstance(log._logger, log._dummyLogger), \
		'main logger wrong instance'
	assert isinstance(log._logger._child, syslog._webLogger), \
		'wrong instance'
	syslog.close()
	assert isinstance(log._logger._child, log._dummyLogger), \
		'wrong instance'
	_cleanup()

def test_initError():
	assert not path.isfile(_dbfile), 'db exists'
	syslog._dbfile = '/i/hope/this/is/invalid/enough/syslog.db'
	syslog.init()
	syslog._dbfile = _dbfile
	assert not path.isfile(_dbfile), 'db was created, so no error happen'

def test_logger():
	l = syslog._logger
	syslog.init()
	assert isinstance(l, syslog._webLogger), 'wrong instance'
	l.error('test.error')
	l.warn('test.warn')
	l.info('test.info')
	l.msg('test.msg')
	assert syslog._getMsgId(0) is None
	msg = syslog._getMsgId(1)
	assert msg[0] == 1
	assert msg[2] == syslog._LVLMAP['error']
	assert msg[3] == 'test.error'
	msg = syslog._getMsgId(2)
	assert msg[0] == 2
	assert msg[2] == syslog._LVLMAP['warn']
	assert msg[3] == 'test.warn'
	msg = syslog._getMsgId(3)
	assert msg[0] == 3
	assert msg[2] == syslog._LVLMAP['info']
	assert msg[3] == 'test.info'
	msg = syslog._getMsgId(4)
	assert msg[0] == 4
	assert msg[2] == syslog._LVLMAP['msg']
	assert msg[3] == 'test.msg'
	syslog.close()
	_cleanup()

def test_lastMsgs():
	l = syslog._logger
	syslog.init()
	assert isinstance(l, syslog._webLogger), 'wrong instance'
	l.error('test.error')
	l.warn('test.warn')
	l.info('test.info')
	l.msg('test.msg')
	assert syslog._getMsgId(0) is None
	msgs = syslog.last(10)
	assert isinstance(msgs, list)
	assert isinstance(msgs[0], Row)
	assert len(msgs) == 4
	idx = 0
	for m in msgs:
		idx += 1
		assert m['id'] == idx
	assert msgs[0]['msg'] == 'test.error'
	assert msgs[1]['msg'] == 'test.warn'
	assert msgs[2]['msg'] == 'test.info'
	assert msgs[3]['msg'] == 'test.msg'
	msgs = syslog.last(2)
	assert len(msgs) == 2
	assert msgs[0]['msg'] == 'test.info'
	assert msgs[1]['msg'] == 'test.msg'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['configure']

def configure(env, cfg):
	env.settings.merge(cfg, 'network.fail2ban', (
		'jail.dir',
		'jail.destdir',
		'jail.disable',
		'jail.enable',
	))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import response

from _sadm import log

__all__ = ['HandlersPlugin']

class HandlersPlugin(object):
	api = 2
	name = 'sadm.listen'

	def setup(self, wapp):
		log.debug('setup')

	def apply(self, callback, ctx):
		if ctx.rule.startswith('/_/'):
			log.debug("apply for rule: %s" % ctx.rule)
			def wrapper(*args, **kwargs):
				log.debug('apply.wrapper')
				resp = callback(*args, **kwargs)
				response.headers['Content-Type'] = 'text/plain; charset=UTF-8'
				return resp
			return wrapper
		log.debug("ignore rule: %s" % ctx.rule)
		return callback



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.errors import Error

def test_error():
	err = Error('errMsg')
	err.typ = 'errType'
	assert str(err) == 'errType: errMsg'
	assert repr(err) == '<sadm.errType: errMsg>'
	err2 = Error('errMsg')
	err2.typ = 'errType'
	assert err == err2
	err2.typ = 'errType2'
	assert err != err2



#!/usr/bin/env python3

# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
from os import system

upload = '--upload' in sys.argv

if upload:
	system('rm -vf dist/sadm-*.*')

system('python3 setup.py clean')
system('python3 setup.py dist')

if upload:
	system('twine upload -s -i jrmsdev@gmail.com dist/sadm-*.*')

sys.exit(0)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import subprocess

from os import environ

from _sadm.errors import CommandError

__all__ = ['call', 'callCheck']

class _ProcMan(object):
	def __init__(self):
		self.call = subprocess.call
		self.check_call = subprocess.check_call

proc = _ProcMan()

def call(cmd, env = None, timeout = None):
	shell = False
	if isinstance(cmd, str):
		shell = True
	return proc.call(cmd, env = env, shell = shell, timeout = timeout)

def callCheck(cmd, env = None):
	if env is None:
		env = environ.copy()
	shell = False
	if isinstance(cmd, str):
		shell = True
	cmderr = None
	try:
		proc.check_call(cmd, env = env, shell = shell)
	except subprocess.CalledProcessError as err:
		cmderr = str(err)
	if cmderr is not None:
		raise CommandError(cmderr)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

# https://docs.pytest.org/en/latest/assert.html

import io
from pytest import raises
from _sadm import log

def test_levels():
	assert sorted(log.levels()) == ['debug', 'error', 'info', 'msg', 'quiet', 'warn']
	assert log.defaultLevel() == 'warn'

def test_dummyLogger():
	log._logger = log._dummyLogger()
	log.debug('')
	log.error('')
	log.warn('')
	log.info('')
	log.msg('')

def _sysLogger(level):
	s = io.StringIO()
	return log._sysLogger(level, outs = s), s

def test_lvlDebug():
	l, s = _sysLogger('debug')
	assert l is not None
	assert s is not None
	l.debug('test')
	assert s.getvalue().startswith('D: ')
	assert s.getvalue().endswith(': test\n')
	l.error('test')
	assert s.getvalue().endswith('\nE: test\n')
	l.warn('test')
	assert s.getvalue().endswith('\nW: test\n')
	l.info('test')
	assert s.getvalue().endswith('\nI: test\n')
	l.msg('test')
	assert s.getvalue().endswith('\ntest\n')

def test_lvlInfo():
	l, s = _sysLogger('info')
	l.debug('test')
	assert '' == s.getvalue()
	l.error('test')
	assert 'E: test\n' == s.getvalue()
	l.info('test')
	assert s.getvalue().endswith('\nI: test\n')
	l.msg('test')
	assert s.getvalue().endswith('\ntest\n')
	l.warn('test')
	assert s.getvalue().endswith('\nW: test\n')

def test_lvlWarn():
	l, s = _sysLogger('warn')
	l.debug('test')
	assert '' == s.getvalue()
	l.info('test')
	assert '' == s.getvalue()
	l.warn('test')
	assert 'W: test\n' == s.getvalue()
	l.error('test')
	assert s.getvalue().endswith('\nE: test\n')
	l.msg('test')
	assert s.getvalue().endswith('\ntest\n')

def test_lvlMsg():
	l, s = _sysLogger('msg')
	l.debug('test')
	assert '' == s.getvalue()
	l.info('test')
	assert '' == s.getvalue()
	l.warn('test')
	assert '' == s.getvalue()
	l.error('test')
	assert 'E: test\n' == s.getvalue()
	l.msg('test')
	assert s.getvalue().endswith('\ntest\n')

def test_lvlError():
	l, s = _sysLogger('error')
	l.debug('test')
	assert '' == s.getvalue()
	l.warn('test')
	assert '' == s.getvalue()
	l.info('test')
	assert '' == s.getvalue()
	l.msg('test')
	assert '' == s.getvalue()
	l.error('test')
	assert 'E: test\n' == s.getvalue()

def test_lvlQuiet():
	l, s = _sysLogger('quiet')
	l.debug('test')
	assert '' == s.getvalue()
	l.warn('test')
	assert '' == s.getvalue()
	l.info('test')
	assert '' == s.getvalue()
	l.msg('test')
	assert '' == s.getvalue()
	l.error('test')
	assert '' == s.getvalue()

def test_lvlInvalid():
	with raises(RuntimeError, match = 'invalid log level: invalid'):
		log._sysLogger('invalid')

def test_debugDepth():
	l, s = _sysLogger('debug')
	l.debug('test', depth = 2)
	msg = s.getvalue()
	assert msg.startswith('D: ')
	assert msg.endswith(': test\n')

def test_debugTag():
	l, s = _sysLogger('debug')
	l.debug('test', tag = 'tag')
	assert s.getvalue().startswith('D: tag ')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import systemd

__all__ = ['build']

def build(env):
	for s in env.settings.sections():
		if s.startswith('docker-compose:'):
			name = s.replace('docker-compose:', '', 1).strip()
			cfg = env.settings[s]
			_composeConfigure(env, name, cfg)

def _composeConfigure(env, name, cfg):
	env.log("docker-compose %s" % name)
	enable = cfg.getboolean('systemd.enable', fallback = True)
	if enable:
		spath = cfg.get('path', fallback = '')
		if spath == '':
			raise env.error("docker-compose:%s path is empty" % name)
		tpldat = {
			'serviceName': name,
			'servicePath': spath,
		}
		systemd.configure(env, 'docker-compose', 'service', name, tpldat)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path

from _sadm.deploy import loader

envfn = path.join('tdata', 'build', 'envsetup', 'build.test.env')
tarfn = path.join('tdata', 'build', 'deploy', 'build.test', 'build.test.tar')
inifn = path.join('tdata', 'build', 'deploy', 'build.test', 'configure.ini')
metafn = path.join('tdata', 'build', 'deploy', 'build.test', 'meta.json')

def test_importenv(env_setup):
	env = env_setup(name = 'build.test', action = 'build')
	assert path.isfile(envfn)
	assert not path.isdir(path.dirname(tarfn))
	loader._importenv(envfn)
	assert path.isfile(tarfn)
	assert path.isfile(inifn)
	assert path.isfile(metafn)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from contextlib import contextmanager

from _sadmtest.mock.utils.cmd import MockCmdProc
from _sadmtest.mock.utils.path import MockPath
from _sadmtest.mock.utils.sh import MockShUtil

import _sadm.utils.cmd
import _sadm.utils.path
import _sadm.utils.sh

@contextmanager
def deploy(name, cfg):
	print('-- mock deploy:', name, cfg)
	try:
		_sadm.utils.cmd.proc = MockCmdProc(cfg)
		_sadm.utils.path._path = MockPath(cfg)
		_sadm.utils.sh.shutil = MockShUtil(cfg)
		yield
		print('-- post check mock deploy:', name)
		_sadm.utils.path._path.check()
		_sadm.utils.sh.shutil.check()
		_sadm.utils.cmd.proc.check()
	finally:
		_sadm.utils.cmd.proc = _sadm.utils.cmd._ProcMan()
		_sadm.utils.path._path = _sadm.utils.path._Path()
		_sadm.utils.sh.shutil = _sadm.utils.sh._ShUtil()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from importlib import import_module
from pytest import raises

from _sadm import configure
from _sadm.errors import PluginError

def test_register_error():
	with raises(RuntimeError, match = 'plugin testing already registered'):
		configure.register('testing', 'filename')

def test_get_plugin():
	p = configure.getPlugin('testing', 'configure')
	assert p.mod == import_module('_sadmtest.plugin.testing.configure')

def test_get_plugin_error():
	with raises(PluginError, match = 'noplugin plugin not found'):
		configure.getPlugin('noplugin', 'configure')
	with raises(PluginError, match = 'testing.nomod: No module named'):
		configure.getPlugin('testing', 'nomod')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import libdir
from _sadm.env.service import Service
from _sadm.utils import path

__all__ = ['configure']

def configure(env, cfg):
	env.settings.merge(cfg, 'service', (
		'config.dir',
		'enable',
	))
	_loadEnabled(env)

def _loadEnabled(env):
	fn = 'service.conf'
	cfgdir = env.settings.get('service', 'config.dir', fallback = 'service')
	for s in env.settings.getlist('service', 'enable'):
		env.log("enable %s" % s)
		libfn = libdir.fpath('services', s, fn)
		libok = _loadConfig(env, libfn)
		sfn = env.assets.rootdir(cfgdir, s, fn)
		sok = _loadConfig(env, sfn)
		if not sok and not libok:
			raise env.error("%s file not found" % sfn)

def _loadConfig(env, fn):
	ok = False
	env.debug("load %s" % fn)
	if path.isfile(fn):
		Service(fn)
		ok = True
	else:
		env.debug("%s file not found" % fn)
	return ok



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path
from pytest import raises

from _sadm import build
from _sadm.errors import BuildError

rootdir = path.join('tdata', 'build', 'envsetup')
builddir = path.join(rootdir, 'build.test')
tarfn = path.join(rootdir, 'build.test.meta', 'build.test.tar')
assetfn = path.join(builddir, 'asset.file')

def build_assets():
	with open(assetfn, 'x') as fh:
		fh.write('testing')

def test_create(env_setup):
	env_setup('build.test', mkdirs = True)
	m = build.Manager(builddir)
	assert not path.isfile(tarfn)
	m.create()
	assert not path.isfile(tarfn)
	assert len(m._data) == 0
	m.close()
	assert path.isfile(tarfn)

def test_create_error(env_setup):
	env_setup('build.test', mkdirs = True)
	m = build.Manager(builddir)
	assert not path.isfile(tarfn)
	with open(tarfn, 'x') as fh:
		fh.write('1')
	assert path.isfile(tarfn)
	with raises(BuildError, match = "%s file exists" % tarfn):
		m.create()

def test_data(env_setup):
	env_setup('build.test', mkdirs = True)
	m = build.Manager(builddir)
	m.create()
	assert len(m._data) == 0
	build_assets()
	m.addfile('asset.file')
	assert len(m._data) == 1
	m.close()

def test_dir_data(env_setup):
	env_setup('build.test', mkdirs = True)
	m = build.Manager(builddir)
	m.create()
	assert len(m._data) == 0
	build_assets()
	m.adddir('.')
	assert len(m._data) == 1
	m.close()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json

from hashlib import sha256
from shutil import make_archive

from _sadm.utils import path, builddir

__all__ = ['pre_build', 'post_build']

_destdir = path.join(path.sep, 'opt', 'sadm')

def pre_build(env):
	env.build.create()

def post_build(env):
	env.build.close()
	_checksum(env)
	_meta(env)
	_zip(env)

def _checksum(env):
	env.log("%s.tar" % env.name())
	fn = builddir.fpath(env, env.name() + '.tar', meta = True)
	h = sha256()
	with open(fn, 'rb') as fh:
		h.update(fh.read())
	env.session.set('sadm.env.checksum', h.hexdigest())

def _zip(env):
	env.log("%s.zip" % env.name())
	rdir = builddir.fpath(env, '.', meta = True)
	fn = builddir.fpath(env, '.')
	make_archive(fn, 'zip', root_dir = rdir, base_dir = '.', verbose = 1)
	h = sha256()
	with open(fn + '.zip', 'rb') as fh:
		h.update(fh.read())
	with open(fn + '.env', 'x') as fh:
		fh.write("%s  %s\n" % (h.hexdigest(),
			path.join(_destdir, 'env', env.name() + '.zip')))

def _meta(env):
	env.log('meta.json')
	with builddir.create(env, 'meta.json', meta = True) as fh:
		json.dump(_getmeta(env), fh, indent = '\t', sort_keys = True)

def _getmeta(env):
	return {
		'sadm.env.name': env.name(),
		'sadm.env.profile': env.profile(),
		'sadm.env.checksum': env.session.get('sadm.env.checksum'),
		'sadm.configure.checksum': env.session.get('sadm.configure.checksum'),
		'sadm.version': env.session.get('sadm.version'),
	}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque
from configparser import ConfigParser, ExtendedInterpolation
from configparser import Error as ParserError

from _sadm import log
from _sadm.configure import pluginsList
from _sadm.errors import SettingsError

__all__ = ['Settings']

_unset = object()

class Settings(ConfigParser):

	def __init__(self):
		super().__init__(defaults = {},
			allow_no_value = False,
			delimiters = ('=',),
			comment_prefixes = ('#',),
			strict = True,
			interpolation = ExtendedInterpolation(),
			default_section = 'default')

	def plugins(self, revert = False):
		for p in pluginsList(revert = revert):
			if self.has_section(p):
				yield p

	def read_file(self, fh):
		log.debug("read file %s" % fh.name)
		try:
			super().read_file(fh)
		except ParserError as err:
			raise SettingsError(str(err))

	def getlist(self, section, option, fallback = _unset):
		try:
			if fallback is _unset:
				s = super().get(section, option)
			else:
				s = super().get(section, option, fallback = '')
				if s == '':
					return fallback
		except ParserError as err:
			raise SettingsError(str(err))
		l = deque()
		for val in s.split():
			val = val.strip()
			for val2 in val.split(','):
				val2 = val2.strip()
				if val2 != '':
					l.append(val2)
		return tuple(l)

	def setlist(self, section, option, args):
		if not self.has_section(section):
			self.add_section(section)
		data = "\n%s" % '\n'.join(args)
		self.set(section, option, data)

	def merge(self, src, section, optfilter):
		if not src.has_section(section):
			return
		if not self.has_section(section):
			self.add_section(section)
		if hasattr(optfilter, '__call__'):
			for opt in src[section]:
				if optfilter(opt) is not None:
					self._overwrite(section, opt, src)
		else:
			for opt in optfilter:
				if opt in src[section]:
					self._overwrite(section, opt, src)

	def _overwrite(self, section, opt, src):
		self[section][opt] = src.get(section, opt, raw = True)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path
from configparser import ConfigParser

from _sadm import log
from _sadm.errors import ProfileError

__all__ = ['Config', 'new']

_cfgFile = path.realpath(path.join('.', 'sadm.cfg'))

_enablePlugins = [
	'sadm',
	'sadmenv',
]

_DEFAULT = {
	'name': '',
	'profile': 'default',
	'env': 'default',
	'dir': '.',
	'plugins': ','.join(_enablePlugins),
}

class Config(ConfigParser):
	_fn = None
	_name = None

	def _init(self, cfgfile = None):
		if cfgfile is None:
			cfgfile = _cfgFile
		self._fn = cfgfile
		log.debug("cfg init %s" % self._fn)
		self.reload()
		n = self.get('default', 'name')
		self._name = n.strip()
		if self._name == '':
			self._name = self._getName()

	def _getName(self):
		return path.basename(path.dirname(self._fn)).strip()

	def name(self):
		return self._name

	def filename(self):
		return self._fn

	def reload(self):
		try:
			with open(self._fn, 'r') as fh:
				self.read_file(fh)
		except FileNotFoundError:
			raise ProfileError("%s file not found" % self._fn)

	def listProfiles(self):
		return sorted(self.sections())

	def listEnvs(self, profile):
		if not self.has_section(profile):
			raise ProfileError("config profile %s not found" % profile)
		e = {}
		for opt in self.options(profile):
			if opt.startswith('env.'):
				n = '.'.join(opt.split('.')[1:]).strip()
				if n != '':
					e[n] = True
		return sorted(e.keys())

	def listPlugins(self, profile):
		return [p.strip() for p in self.get(profile, 'plugins').split(',')]

def new(cfgfile = None):
	log.debug('cfg.new')
	config = Config(
		defaults = _DEFAULT,
		strict = True,
		default_section = 'default',
		empty_lines_in_values = False,
		comment_prefixes = ('#', ),
		delimiters = ('=', ),
	)
	config._init(cfgfile = cfgfile)
	return config



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from io import StringIO
from pytest import raises

from _sadm.errors import SessionError
from _sadm.env.session import Session

def test_session():
	sess = Session()
	assert isinstance(sess._d, dict)
	assert sess._start is None
	sess.start()
	assert sess._start is not None
	took = sess.stop()
	assert sess._start is None
	assert isinstance(took, float)

def test_start_error():
	sess = Session()
	sess.start()
	with raises(SessionError, match = 'session already started'):
		sess.start()

def test_stop_error():
	sess = Session()
	sess.start()
	sess.stop()
	sess._done = False
	with raises(SessionError, match = 'session not started'):
		sess.stop()

def test_set_get():
	sess = Session()
	sess.start()
	sess.set('test', 'testing')
	assert sess.get('test') == 'testing'
	sess.stop()

def test_get_default():
	sess = Session()
	sess.start()
	assert sess.get('test', 'testing_default') == 'testing_default'
	sess.stop()

def test_set_get_error():
	sess = Session()
	sess.start()
	with raises(SessionError, match = 'test option not set'):
		assert sess.get('test') == 'testing'
	sess.set('test', 'testing')
	with raises(SessionError, match = 'test option already set'):
		sess.set('test', 'testing')
	sess.stop()

def test_dump():
	sess = Session()
	sess.start()
	fh = StringIO()
	sess.dump(fh, indent = None)
	fh.seek(0, 0)
	assert fh.read() == '{}'
	fh.truncate()
	fh.seek(0, 0)
	sess.set('opt1', 'val1')
	sess.set('opt2', 'val2')
	sess.dump(fh, indent = None)
	fh.seek(0, 0)
	assert fh.read() == '{"opt1": "val1", "opt2": "val2"}'
	sess.stop()

def test_stopped_error():
	sess = Session()
	sess.start()
	sess.stop()
	with raises(SessionError, match = 'session stopped'):
		sess.set('testing', 'stopped')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log
from _sadm.deploy.loader import loadenv

def cmdArgs(parser):
	p = parser.add_parser('import', help = 'import sadm.env')
	p.add_argument('filename', help = 'sadm.env file to import')
	p.set_defaults(command = 'import')

def main(args):
	log.debug("import %s" % args.filename)
	return loadenv(args.filename)



#!/usr/bin/env python3

# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys

from _sadm import listen, libdir
from _sadm.cmd import flags
from _sadm.utils import path
from _sadm.utils.cmd import callCheck

def bottle():
	return listen.start()

def uwsgi():
	cmd = [
		'uwsgi',
		'--need-plugin', 'python3',
		'--pyhome', sys.exec_prefix,
		'--set-ph', 'netloc=127.0.0.1:3666',
		'--touch-reload', path.join(path.sep, 'run', 'sadm.listen.uwsgi.reload'),
		'--ini', libdir.fpath('listen', 'uwsgi.ini'),
	]
	callCheck(cmd)
	return 0

if __name__ == '__main__':
	if '--bottle' in sys.argv:
		sys.exit(bottle())
	else:
		sys.exit(uwsgi())



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from collections import deque

__all__ = ['cfgfilter', 'configure']

def cfgfilter(env, pcfg, dep):
	dn = env.dist() + '.'
	def pfilter(opt):
		if opt.startswith(dn):
			return opt
		return None
	env.settings.merge(pcfg, dep, pfilter)

def configure(env, cfg):
	data = deque()
	for opt in cfg['sync']:
		data.append(getInfo(cfg, opt))
	env.session.set('sync', tuple(data))

def getInfo(cfg, opt):
	i = cfg.getlist('sync', opt)
	inf = {'cfg.opt': opt, 'src': i[0], 'dst': i[1], 'args': dict()}
	if len(i) > 2:
		for arg in i[2:]:
			if arg.startswith('user:'):
				val = arg.split(':')[1]
				inf['args']['user'] = val
			elif arg.startswith('user='):
				val = arg.split('=')[1]
				inf['args']['user'] = val

			elif arg.startswith('group:'):
				val = arg.split(':')[1]
				inf['args']['group'] = val
			elif arg.startswith('group='):
				val = arg.split('=')[1]
				inf['args']['group'] = val

			elif arg.startswith('filemode:'):
				val = arg.split(':')[1]
				inf['args']['filemode'] = val
			elif arg.startswith('filemode='):
				val = arg.split('=')[1]
				inf['args']['filemode'] = val

			elif arg.startswith('dirmode:'):
				val = arg.split(':')[1]
				inf['args']['dirmode'] = val
			elif arg.startswith('dirmode='):
				val = arg.split('=')[1]
				inf['args']['dirmode'] = val
	return inf



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log
from _sadm.utils import sh
from _sadm.utils.cmd import callCheck

__all__ = ['GitRepo']

class GitRepo(object):

	def hook(self, action, args):
		repodir = args.get('repo.path', 'NOREPOPATH')
		log.debug("hook action %s repo dir %s" % (action, repodir))
		sh.chdir(repodir)
		if action == 'push':
			self._pull()
			# TODO: vcs.repo.deploy

	def _pull(self):
		log.debug('git pull')
		callCheck(['git', 'pull'])



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import libdir
from _sadm.utils import builddir, path
from _sadm.utils.cmd import call, callCheck

__all__ = ['status', 'start', 'stop', 'restart', 'reload', 'enable', 'configure']

_ctlcmd = ['systemctl', '--quiet', '--no-pager', '--no-legend', '--no-ask-password']

def _cmd(service, action, *args, check = True):
	cmd = []
	cmd.extend(_ctlcmd)
	cmd.extend(args)
	cmd.append(action)
	cmd.append(service)
	if check:
		return callCheck(cmd)
	return call(' '.join(cmd))

def status(service, *args):
	return _cmd(service, 'status', *args, check = False)

def start(service, *args):
	return _cmd(service, 'start', *args)

def stop(service, *args):
	return _cmd(service, 'stop', *args)

def restart(service, *args):
	return _cmd(service, 'restart', *args)

def reload(service, *args):
	return _cmd(service, 'reload', *args)

def enable(service, *args):
	return _cmd(service, 'enable', *args)

def configure(env, unit, typ, name, tpldat):
	service = "%s-%s.%s" % (unit, name, typ)
	with libdir.fopen('utils', 'systemd', "%s.%s" % (unit, typ)) as fh:
		data = fh.read()
	dst = path.join('etc', 'systemd', 'system', service)
	with builddir.create(env, dst) as fh:
		fh.write(data.format(**tpldat))
		fh.flush()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log, cfg
from _sadm.web import tpl
from _sadm.web.app import wapp, view

@wapp.route('/profile')
@view('profile.html')
@tpl.data('profile')
def index():
	log.debug('index')
	return {
		'profiles': _getallProfiles(),
	}

def _getallProfiles():
	config = cfg.new()
	l = []
	for p in config.listProfiles():
		l.append(_getProfile(p, config))
	return l

def _getProfile(p, config):
	return {
		'name': p,
		'envs': config.listEnvs(p),
	}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import json
import sys
from urllib.parse import urlparse

from _sadm import log, libdir, version
from _sadm.listen.client import ListenClient
from _sadm.utils import sh, path
from _sadm.utils.cmd import callCheck

__all__ = ['dispatch', 'main']

#
# webhook dispatch task (exec new process)
#

def dispatch(req, task, action, taskArgs):
	log.debug("dispatch task: %s - action: %s" % (task, action))
	reqURL = urlparse(req.url)
	if reqURL.scheme == '':
		reqURL.scheme = 'http'
	obj = {
		'sadm.log': log.curLevel(),
		'sadm.listen.url': "%s://%s" % (reqURL.scheme, reqURL.netloc),
		'task': task,
		'task.action': action,
		'task.args': taskArgs,
	}
	taskfn = None
	with sh.mktmp(prefix = __name__, suffix = ".%s.json" % task) as fh:
		taskfn = fh.name()
		fh.write(json.dumps(obj))
	_sched(taskfn)

def _sched(taskfn):
	self = libdir.fpath('listen', 'exec.py')
	cmd = [path.join(sys.exec_prefix, 'bin', 'python3'), self, taskfn]
	atcmd = "echo 'sleep 1 && %s' | at now" % ' '.join(cmd)
	log.debug("run: %s" % atcmd)
	callCheck(atcmd)

#
# task exec main (runs under a new process)
#

def main(args):
	if len(args) != 1:
		print('ERROR: sadm-listen exec invalid args', file = sys.stderr)
		return 1
	taskfn = args[0]
	obj = None
	try:
		with open(taskfn, 'r') as fh:
			obj = json.load(fh)
	except Exception as err:
		log.error("%s" % err)
		return 2
	finally:
		path.unlink(taskfn)
	log.init(obj.get('sadm.log', 'warn'))
	log.debug(version.string('sadm-listen'))
	log.debug("task file %s" % taskfn)
	task = obj.get('task', None)
	if task is None:
		log.error('listen.exec task not set')
		return 3
	taskAction = obj.get('task.action', None)
	if taskAction is None:
		log.error("listen.exec task %s: no action" % task)
		return 4
	taskArgs = obj.get('task.args', None)
	if taskArgs is None:
		log.error("listen.exec task %s: no args" % task)
		return 5
	cliURL = obj.get('sadm.listen.url', 'http://127.0.0.1:3666')
	cli = ListenClient(cliURL)
	cli.exec(task, taskAction, taskArgs)
	return 0

if __name__ == '__main__':
	sys.exit(main(sys.argv[1:]))



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log
from _sadm.web import tpl
from _sadm.web.app import wapp, view

@wapp.error(500)
@view('error.html')
@tpl.data('err500')
def err500(err):
	log.debug("error type %s" % type(err))
	log.debug("%s - %s" % (err.status_line, err.exception))
	return {
		'err': err,
	}

@wapp.error(404)
@view('error.html')
@tpl.data('err400')
def err400(err):
	log.debug(err.status_line)
	return {
		'err': err,
	}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from os import path
from pytest import raises

from _sadm.configure import plugins
from _sadm.env.settings import Settings
from _sadm.errors import EnvError, AssetNotFoundError

def test_configure(testing_env):
	e = testing_env()
	e.session.start()
	plugins.configure(e)
	assert isinstance(e.settings, Settings)
	assert sorted(e.settings.sections()) == ['os', 'sadm', 'sync', 'testing']

def test_file_not_found(testing_env):
	e = testing_env()
	with raises(AssetNotFoundError):
		plugins.configure(e, cfgfile = 'nofile.ini')

def test_cfg_invalid_name(testing_env):
	e = testing_env()
	with raises(EnvError, match = 'invalid config name'):
		plugins.configure(e, cfgfile = 'testing/config-invalid-name.ini')

def test_deploy_mode(testing_env):
	e = testing_env()
	e.session.start()
	plugins.configure(e, cfgfile = path.join('testing', 'deploy-configure.ini'))
	assert e.settings.sections() == ['sadmenv']



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
from _sadm import log, env
from _sadm.cmd import flags
from _sadm.cmd.deploy import loader
from _sadm.cmd.deploy import deploy as deploy_cmd

_parser = flags.new('sadm', desc = 'deploy sadm env')

def _getArgs():
	subparser = _parser.add_subparsers(title = 'commands',
		description = 'run `sadm command -h` for more information',
		help = 'description')
	loader.cmdArgs(subparser)
	deploy_cmd.cmdArgs(subparser)
	return flags.parse(_parser)

def main():
	args = _getArgs()
	log.debug("deploy %s/%s" % (args.profile, args.env))
	try:
		cmd = args.command
	except AttributeError:
		log.error('invalid usage')
		_parser.print_usage()
		return 1
	log.debug("dispatch command %s" % cmd)
	if args.command == 'import':
		return loader.main(args)
	return deploy_cmd.main(args)

if __name__ == '__main__':
	sys.exit(main())
